from typing import Dict, List, Optional, Tuple, Union

from shapely.geometry import LinearRing, LineString, Polygon


def load_road_geometry_from_json(
    data: Dict,
) -> Tuple[Optional[Polygon], Optional[LineString]]:
    """Load the boundary and center from json data."""
    if "Boundary" in data:
        boundary = data["Boundary"]
        if isinstance(boundary, list):
            boundary = Polygon([[v["x"], v["y"]] for v in boundary])
        elif isinstance(boundary, dict):
            boundary = Polygon(
                [[v["x"], v["y"]] for v in boundary["exterior"]],
                holes=[
                    LinearRing([[v["x"], v["y"]] for v in i])
                    for i in boundary["interiors"]
                ],
            )
        else:
            raise ValueError(
                f"Type {type(boundary)} is not supported for boundary."
            )
    else:
        boundary = None
    if "Center" in data:
        center = data["Center"]
        if isinstance(center, list):
            center = LineString([[v["x"], v["y"]] for v in center])
        else:
            raise ValueError(f"Type {type(center)} is not supported for center.")
    else:
        center = None
    return boundary, center


def polygon_to_data(
    poly: Polygon,
) -> Union[List[Dict[str, float]], Dict[str, List[Dict[str, float]]]]:
    """Return a dict or list representing the polygon data."""
    exterior = [{"x": float(x), "y": float(y)} for x, y in poly.exterior.coords]
    if not poly.interiors:
        return exterior
    return {
        "exterior": exterior,
        "interiors": [
            [{"x": float(x), "y": float(y)} for x, y in i.coords]
            for i in poly.interiors
        ],
    }



import os
from copy import deepcopy
from tempfile import TemporaryDirectory

import numpy as np
from lxml import etree

from scenario_gym.catalog_entry import Catalog
from scenario_gym.entity import Pedestrian, Vehicle
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.xosc_interface import (
    import_scenario,
    read_catalog,
    write_catalogs,
    write_scenario,
)


def test_read_catalog(all_catalogs):
    """Test reading catalogs."""

    def test_entry(e):
        """Check the attributes of a loaded entity."""
        if isinstance(e, Vehicle):
            assert e.catalog_entry.rear_axle.wheel_diameter > 0
            assert e.catalog_entry.mass > 0
        elif isinstance(e, Pedestrian):
            assert e.catalog_entry.mass > 0
        else:
            raise TypeError(f"No catalog type for {e}.")

    for catalog in all_catalogs.values():
        if "Scenario_Gym" not in catalog:
            continue
        _, entries = read_catalog(catalog)
        for e in entries.values():
            try:
                test_entry(e)
            except TypeError as err:
                print(catalog)
                raise err


def test_import_scenario(all_scenarios):
    """Test importing a scenario."""
    path = all_scenarios["511c3b32-1bd5-4c16-9649-4c92da443301"]
    scenario = import_scenario(path)

    assert len(scenario.entities) == 3
    assert "Greenwich_Road_Network_002" in scenario.road_network.name
    assert scenario.entities[0].ref == "ego"
    assert isinstance(scenario.entities[0], Vehicle)
    assert isinstance(scenario.entities[2], Pedestrian)
    assert scenario.entities[0].catalog_entry.rear_axle.wheel_diameter > 0
    assert scenario.entities[2].catalog_entry.mass > 0


def test_write_scenario(all_scenarios) -> None:
    """
    Rollout a single scenario and write to a new scenario.

    Output the xosc then load it again and rollout the
    recorded version.

    """
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    out_path = scenario_path.replace("Scenarios", "Recordings").replace(
        ".xosc", "_test.xosc"
    )
    scenario = import_scenario(scenario_path).reset_start()

    # rollout
    gym = ScenarioGym()
    gym.set_scenario(scenario)
    gym.rollout()
    old_scenario = gym.state.scenario
    traj1 = old_scenario.entities[0].trajectory

    # output to OpenSCENARIO
    new_scenario = gym.state.to_scenario()
    new_scenario.properties["a"] = 1
    new_scenario.properties["files"] = ["a.txt"]
    write_scenario(new_scenario, out_path)

    # reload and test
    n_entities = len(gym.state.scenario.entities)
    n_stationary = sum(
        1 for t in gym.state.scenario.trajectories.values() if len(t) == 1
    )
    gym.load_scenario(out_path)
    assert gym.state.scenario.properties["a"] == 1, "Properties not copied."
    assert gym.state.scenario.properties["files"] == ["a.txt"]
    traj2 = gym.state.scenario.entities[0].trajectory
    assert (
        len(gym.state.scenario.entities) == n_entities
    ), "New scenario has a different number of entities."
    assert all(
        (
            isinstance(entity, type(old_entity))
            for entity, old_entity in zip(
                old_scenario.entities, gym.state.scenario.entities
            )
        )
    ), "Entities are not the same type."
    assert n_stationary == sum(
        1 for t in gym.state.scenario.trajectories.values() if len(t) == 1
    ), "New scenario has a different number of stationary entities."
    assert all(
        [
            np.allclose(traj1.position_at_t(0.0), traj2.position_at_t(0.0)),
            np.allclose(traj1.position_at_t(5.0), traj2.position_at_t(5.0)),
            np.allclose(traj1.position_at_t(10.0), traj2.position_at_t(10.0)),
        ]
    ), "Recorded and true trajectories differ."


def test_properties(all_scenarios):
    """Test loading properties from xosc."""
    s = all_scenarios["3e39a079-5653-440c-bcbe-24dc9f6bf0e6"]
    s = import_scenario(s)
    assert s.properties["prop"] == 64, "Property not loaded correctly."
    assert s.properties["prop2"] == "64a", "Property not loaded correctly."
    assert s.properties["files"] == ["test.txt"], "Property not loaded correctly."


def test_write_scenario_without_references(all_scenarios) -> None:
    """
    Rollout a single scenario and write to a new scenario.

    Output the xosc then load it again and rollout the
    recorded version.

    """
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    out_path = scenario_path.replace("Scenarios", "Recordings").replace(
        ".xosc", "_test_no_refs.xosc"
    )
    scenario = import_scenario(scenario_path).reset_start()

    # rollout
    gym = ScenarioGym()
    gym.set_scenario(scenario)
    gym.rollout()

    # output to OpenSCENARIO
    new_scenario = gym.state.to_scenario()
    write_scenario(new_scenario, out_path, use_catalog_references=False)

    et = etree.parse(out_path)
    assert len(et.getroot().findall("Entities/ScenarioObject/Vehicle")) == len(
        new_scenario.vehicles
    ), "Not all vehicles were written."
    assert (
        len(et.getroot().findall("Entities/ScenarioObject/CatalogReference")) == 0
    ), "Catalog references were written."


def test_write_catalogs(all_scenarios):
    """Test creating catalogs."""
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    scenario = import_scenario(scenario_path)
    entries = [deepcopy(scenario.entities[0].catalog_entry)]

    new_catalog = Catalog(
        entries[0].catalog.name + "_test",
        "Test_Scenario_Gym",
    )
    for ce in entries:
        ce.catalog = new_catalog

    with TemporaryDirectory() as tmpdir:
        write_catalogs(tmpdir, entries)

        assert os.path.isdir(
            os.path.join(tmpdir, "Test_Scenario_Gym")
        ), "Catalog directory not created."
        assert os.path.isdir(
            os.path.join(tmpdir, "Test_Scenario_Gym", "VehicleCatalogs")
        ), "Vehicle catalog directory not created."

        cat_path = os.path.join(
            tmpdir,
            "Test_Scenario_Gym",
            "VehicleCatalogs",
            f"{entries[0].catalog.name}.xosc",
        )
        assert os.path.isfile(cat_path), "Catalog not written."

        read_cat, entries2 = read_catalog(cat_path)

    assert read_cat.group_name == "Test_Scenario_Gym", "Catalog not written."
    assert (
        entries2["car1"].catalog_entry.mass == entries[0].mass
    ), "Catalog entry not written."



"""
Import scenarios from the Argoverse 2 Motion Forecasting dataset.

https://www.argoverse.org/about.html#terms-of-use
"""
import json
from contextlib import suppress
from pathlib import Path
from typing import Dict

import numpy as np
from shapely.geometry import LineString, Polygon

from scenario_gym.catalog_entry import BoundingBox, Catalog, CatalogEntry
from scenario_gym.entity import Entity
from scenario_gym.road_network import (
    Lane,
    LaneType,
    Road,
    RoadGeometry,
    RoadNetwork,
)
from scenario_gym.scenario import Scenario
from scenario_gym.trajectory import Trajectory

try:
    import pandas as pd
except ImportError:
    raise ImportError(
        """
\tPandas is required for this integration.
\tInstall by `pip install pandas`.
"""
    )


class Lane(Lane):
    """Add the argoverse information to the lane object."""

    def __init__(
        self,
        is_intersection: bool,
        left_neighbour_id: str,
        right_neighbour_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.is_intersection = is_intersection
        self.left_neighbour_id = left_neighbour_id
        self.right_neighbour_id = right_neighbour_id


track_types = [
    "VEHICLE",
    "PEDESTRIAN",
    "MOTORCYCLIST",
    "CYCLIST",
    "BUS",
    "STATIC",
    "BACKGROUND",
    "CONSTRUCTION",
    "RIDERLESS_BICYCLE",
    "UNKNOWN",
]


class Catalogs:
    """
    Catalogs used in the argoverse dataset.

    Using the dimensions defined in their scenario_visualization.py
    """

    vehicle_box = BoundingBox(1.8, 3.8, 0.0, 0.0)
    argoverse_catalog = Catalog("ArgoverseCatalog", None)
    vehicle = CatalogEntry(
        argoverse_catalog,
        "vehicle",
        "car",
        "Vehicle",
        vehicle_box,
    )

    pedestrian_box = BoundingBox(0.4, 0.4, 0.0, 0.0)
    pedestrian = CatalogEntry(
        argoverse_catalog,
        "pedestrian",
        "pedestrian",
        "Pedestrian",
        pedestrian_box,
    )

    motorbike_box = BoundingBox(0.2, 0.8, 0.0, 0.0)
    motorcyclist = CatalogEntry(
        argoverse_catalog,
        "motorcyclist",
        "motorbike",
        "Vehicle",
        motorbike_box,
    )

    cyclist_box = BoundingBox(0.7, 2.0, 0.0, 0.0)
    cyclist = CatalogEntry(
        "ArgoverseCatalog",
        "cyclist",
        "bicycle",
        "Vehicle",
        cyclist_box,
    )

    bus_box = BoundingBox(2.8, 11.0, 0.0, 0.0)
    bus = CatalogEntry(
        argoverse_catalog,
        "bus",
        "bus",
        "Vehicle",
        bus_box,
    )

    riderless_bicycle_box = BoundingBox(0.3, 1.5, 0.0, 0.0)
    riderless_bicycle = CatalogEntry(
        argoverse_catalog,
        "riderless_bicycle",
        "obstacle",
        "Vehicle",
        riderless_bicycle_box,
    )


def import_argoverse_scenario(path: str) -> Scenario:
    """
    Import a recorded scenario from the argoverse data.

    This assumes fixed bounding box sizes for each entity. For now
    ignoring object types: background, construction,
    static and unkown.
    """
    path = Path(path)
    scenario_id = path.parts[-1]

    pq_path = Path(path, f"scenario_{scenario_id}.parquet")
    main_df = pd.read_parquet(pq_path).sort_values("timestep")
    dfs = list(main_df.groupby("track_id"))
    all_ids = sorted(main_df["track_id"].unique())
    assert "AV" in all_ids, "No AV found to use as ego."
    all_ids.remove("AV")

    entities = []
    for track_id, df in dfs:

        if track_id != "AV" and not df["observed"].any():
            continue

        # get catalog
        object_type = df["object_type"].iloc[0]
        with suppress(AttributeError):
            catalog_entry = getattr(Catalogs, object_type)

        # get start and end in seconds
        start = df["start_timestamp"].iloc[0] / 1e9
        end = df["end_timestamp"].iloc[0] / 1e9
        num = df["num_timestamps"].iloc[0] - 1
        t_scale = (end - start) / num

        # build trajectory
        traj_data = df[
            [
                "timestep",
                "position_x",
                "position_y",
                "heading",
            ]
        ].to_numpy()
        traj_data[:, 0] = t_scale * traj_data[:, 0]

        v0 = df[["velocity_x", "velocity_y"]].iloc[0].to_numpy()
        t_pre = np.array(
            [
                -0.1,
                *(traj_data[0, [1, 2]] - 0.1 * v0),
                traj_data[0, 3],
            ]
        )

        traj_data = np.concatenate(
            [
                t_pre[None],
                traj_data,
            ],
            axis=0,
        )
        trajectory = Trajectory(traj_data, fields=["t", "x", "y", "h"])

        entity_ref = (
            f"entity_{1+all_ids.index(track_id)}" if track_id != "AV" else "ego"
        )
        entity = Entity(catalog_entry, ref=entity_ref)
        entity.trajectory = trajectory
        entities.append(entity)

    ego = None
    for e in entities:
        if e.ref == "ego":
            ego = e
            break
    if ego is not None:
        entities.remove(ego)
        entities.insert(0, ego)

    road_network_data = json.load(
        open(Path(path, f"log_map_archive_{scenario_id}.json"), "r")
    )
    road_network = create_argoverse_road_network(road_network_data)

    scenario = Scenario(
        entities,
        name=scenario_id,
        path=str(path.absolute()),
        road_network=road_network,
    )
    return scenario


def create_argoverse_road_network(data: Dict) -> RoadNetwork:
    """Create a road network from the argoverse log map."""
    driveable_areas = []
    for area in data["drivable_areas"].values():
        poly = Polygon([[v["x"], v["y"]] for v in area["area_boundary"]])
        driveable_areas.append(
            RoadGeometry(
                area["id"],
                poly,
            )
        )

    roads = []
    for l_data in data["lane_segments"].values():
        center = LineString([[d["x"], d["y"]] for d in l_data["centerline"]])
        boundary = center.buffer(1.75, cap_style=2)
        lane = Lane(
            l_data["is_intersection"],
            l_data["left_neighbor_id"],
            l_data["right_neighbor_id"],
            l_data["id"],
            boundary,
            center,
            l_data["successors"],
            l_data["predecessors"],
            LaneType.driving,  # data["lane_type"],
        )
        roads.append(
            Road(
                f"road_{l_data['id']}",
                boundary,
                center,
                [lane],
            )
        )

    # TODO pedestrian crossings
    return RoadNetwork(
        roads=roads,
        intersections=[],
        driveable_areas=driveable_areas,
    )



from typing import Optional

from scenario_gym.action import Action, TeleportAction
from scenario_gym.controller import (
    Controller,
    PIDController,
    ReplayTrajectoryController,
)
from scenario_gym.entity import Entity
from scenario_gym.observation import Observation
from scenario_gym.scenario import Scenario
from scenario_gym.sensor import EgoLocalizationSensor, Sensor
from scenario_gym.state import State
from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import ArrayLike


class Agent:
    """Base agent class. Processes observations to select an action."""

    def __init__(self, entity: Entity, controller: Controller, sensor: Sensor):
        """
        Construct an agent given an entity, controller and sensor.

        Parameters
        ----------
        entity : Entity
            The entity that the agent will control.

        controller : Controller
            The controller for the agent.

        sensor : Sensor:
            The sensor module for the agent.

        """
        self.entity = entity
        self.controller = controller
        self.sensor = sensor
        self._last_action: Optional[Action] = None
        self._last_reward: Optional[float] = None
        self._trajectory: Optional[Trajectory] = None

    def reset(self, state: State) -> None:
        """Reset the agent state at the start of the scenario."""
        self.last_action = None
        self.last_reward = None
        self.sensor.reset(state)
        self.controller.reset(state)
        self._reset()

    def step(self, state: State) -> ArrayLike:
        """Select an action from the current observation."""
        obs = self.sensor.step(state)
        action = self._step(obs)
        self.last_action = action
        return self.controller.step(state, action)

    def _reset(self) -> None:
        """Reset the agent state at the start of the scenario."""
        pass

    def _step(self, observation: Observation) -> Action:
        """Select an action from the current observation."""
        pass

    def finish(self, state: State) -> None:
        """Process the end of the scenario."""
        pass

    @property
    def trajectory(self) -> Trajectory:
        """
        Return the trajectory of the agent.

        By default this is the underlying entities trajectory but can be overridden.
        """
        return (
            self._trajectory
            if self._trajectory is not None
            else self.entity.trajectory
        )

    @trajectory.setter
    def trajectory(self, trajectory: Trajectory):
        self._trajectory = trajectory

    @property
    def last_action(self) -> Action:
        """Return the previous action selected by the agent."""
        return self._last_action

    @last_action.setter
    def last_action(self, action: Action) -> None:
        self._last_action = action

    def reward(self, state: State) -> Optional[float]:
        """Return and cache the reward for the agent from the current state."""
        r = self._reward(state)
        if r is not None:
            self.last_reward = r
        return r

    def _reward(self, state: State) -> Optional[float]:
        """Return the reward for the agent from the current state."""
        pass

    @property
    def last_reward(self) -> Optional[float]:
        """Get the last reward."""
        return self._last_reward

    @last_reward.setter
    def last_reward(self, reward: Optional[float]) -> None:
        self._last_reward = reward


class ReplayTrajectoryAgent(Agent):
    """Agent for entities that follow trajectories predefined in the scenario."""

    def _reset(self) -> None:
        """Reset the agent state at the start of the scenario."""
        pass

    def _step(self, observation: Observation) -> Action:
        """Return the agent's next pose."""
        new_pose = self.trajectory.position_at_t(observation.next_t)
        return TeleportAction(pose=new_pose)


class PIDAgent(Agent):
    """Agent following a specified trajectory with a PID controller."""

    def __init__(self, entity: Entity, **controller_kwargs):
        super().__init__(
            entity,
            PIDController(entity, **controller_kwargs),
            EgoLocalizationSensor(entity),
        )

    def _reset(self) -> None:
        """Reset the agent state at the start of the scenario."""
        pass

    def _step(self, observation: Observation) -> TeleportAction:
        """Get the next waypoint from the agent's trajectory."""
        pos = self.trajectory.position_at_t(observation.next_t)
        return TeleportAction(x=pos[0], y=pos[1], z=pos[2])


def _create_agent(scenario: Scenario, entity: Entity) -> Optional[Agent]:
    """
    Return a replay trajectory agent.

    This is the default create agent function used by the gym.

    Parameters
    ----------
    scenario : Scenario
        The scenario object.

    entity : Entity
        The specific entity within the scenario.

    """
    if entity.ref == "ego":
        controller = ReplayTrajectoryController(entity)
        sensor = EgoLocalizationSensor(entity)
        return ReplayTrajectoryAgent(entity, controller, sensor)



from __future__ import annotations

import warnings
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np
from shapely.geometry import MultiPolygon, Point, Polygon
from shapely.vectorized import contains

from scenario_gym.callback import StateCallback
from scenario_gym.entity import BatchReplayEntity, Entity
from scenario_gym.road_network import RoadObject
from scenario_gym.scenario import Scenario, ScenarioAction
from scenario_gym.state.utils import detect_collisions
from scenario_gym.trajectory import Trajectory, is_stationary

Agent = TypeVar("Agent")


class State:
    """
    The global state of the gym.

    Holds the current time, the terminal state and the positions and velocities of
    all the entities.

    Can also be parameterised with different end conditions for the scenario.
    E.g. to end when the recorded scenario ends or if a collision occurs.
    Additional information may be provided through custom methods passed
    as state_callbacks.
    """

    def __init__(
        self,
        scenario: Scenario,
        scenario_path: Optional[str] = None,
        persist: bool = False,
        conditions: Optional[List[Union[str, Callable[[State], bool]]]] = None,
        state_callbacks: Optional[Dict[str, StateCallback]] = None,
    ):
        """
        Init the state.

        Parameters
        ----------
        scenario : Scenario
            The scenario to be simulated.

        scenario_path : Optional[str]
            The path to the scenario file if loaded from one.

        persist : bool
            Whether entities should persist in the simulation.

        conditions : Optional[List[Union[str, Callable[[State], bool]]]]
            Terminal conditions that will end the scenario if any is met. May be a
            string referencing an entry of the TERMINAL_CONDITIONS dictionary.

        state_callbacks : Optional[List[StateCallback]]
            Methods to be called on the state when the timestep is updated.
            Can be used to add additional information to the state that can is then
            accessible by all agents.

        """
        self._scenario = scenario
        self.scenario_path = scenario_path
        self.persist = persist
        if conditions is None:
            self.terminal_conditions = [TERMINAL_CONDITIONS["max_length"]]
        else:
            self.terminal_conditions = [
                cond if callable(cond) else TERMINAL_CONDITIONS[cond]
                for cond in conditions
            ]
        self.state_callbacks = [] if state_callbacks is None else state_callbacks

        self.next_t: Optional[float] = None
        self._t: Optional[float] = None
        self._prev_t: Optional[float] = None
        self.is_done = False
        self.last_keystroke: Optional[int] = None

        self._collisions: Optional[Dict[Entity, List[Entity]]] = None
        self._callbacks: Dict[Type[StateCallback], StateCallback] = {}

        self.unapplied_actions: List[ScenarioAction]
        self.action_apply_times: Dict[ScenarioAction, float]

        self.all_entities: List[Entity]
        self.poses: Dict[Entity, np.ndarray]
        self.prev_poses: Dict[Entity, np.ndarray]
        self.velocities: Dict[Entity, np.ndarray]
        self.distances: Dict[Entity, float]
        self.entity_state: Dict[Entity, Any]
        self._recorded_poses: Dict[Entity, List[Tuple[float, np.ndarray]]]

        self.agents: Dict[Entity, Agent] = {}
        self.non_agents = BatchReplayEntity(persist=persist)

    @property
    def scenario(self) -> Scenario:
        """Get the current scenario."""
        return self._scenario

    def reset(self, t_minus1: float, t_0: float) -> None:
        """
        Reset the state to the initial timestep.

        Parameters
        ----------
        t_minus1 : float
            Time before the initial timestep to use for initial velocities.

        t_0 : float
            Initial timestep.

        """
        self._reset_data()
        self.is_done = False

        # set initial poses
        # always use extrapolation for previous poses to get the initial velocity
        prev_poses, poses = {}, {}
        for entity in self.all_entities:
            pose = entity.trajectory.position_at_t(
                t_0,
                extrapolate=(
                    entity.is_static()
                    or ((False, False) if self.persist else False)
                ),
            )
            if pose is not None:
                poses[entity] = pose
                prev_poses[entity] = entity.trajectory.position_at_t(
                    t_minus1, extrapolate=((False, False) if self.persist else True)
                )
        self.update_poses(t_minus1, prev_poses)
        self.update_poses(t_0, poses)
        self.update_actions()

        for cb in self.state_callbacks:
            cb.reset(self)
        self.update_callbacks()

        for agent in self.agents.values():
            agent.reset(self)

    def _reset_data(self) -> None:
        """Reset stored simulation data."""
        self.next_t: Optional[float] = None
        self._t: Optional[float] = None
        self._prev_t: Optional[float] = None
        self.unapplied_actions = self.scenario.actions.copy()
        self.action_apply_times = {
            a: float("nan") for a in self.scenario.actions.copy()
        }

        self.all_entities = self.scenario.entities.copy()
        self.poses: Dict[Entity, np.ndarray] = {}
        self.prev_poses: Dict[Entity, np.ndarray] = {}
        self.velocities: Dict[Entity, np.ndarray] = {}
        self.distances: Dict[Entity, float] = dict.fromkeys(self.all_entities, 0.0)
        self.entity_state: Dict[Entity, Any] = dict.fromkeys(self.all_entities)
        self._recorded_poses: Dict[Entity, List[Tuple[float, np.ndarray]]] = {}
        for entity in self.all_entities:
            self._recorded_poses[entity] = []

    def step(self, new_poses: Dict[Entity, np.ndarray]) -> None:
        """Update by one timestep."""
        self._clear_cache()
        self.update_poses(self.next_t, new_poses.copy())
        self.update_actions()
        self.update_callbacks()
        self.is_done = self.check_terminal()

    def _clear_cache(self) -> None:
        """Clear cached data on step."""
        self._collisions = None
        self._callbacks = {}

    @property
    def t(self):
        """Get the time in seconds (s)."""
        return self._t

    @t.setter
    def t(self, t: float) -> None:
        self.prev_t = self._t
        self._t = t
        return self._t

    @property
    def prev_t(self) -> float:
        """Get the previous time (s)."""
        return self._prev_t

    @prev_t.setter
    def prev_t(self, prev_t: float) -> None:
        self._prev_t = prev_t

    @property
    def dt(self) -> float:
        """Get the previous timestep."""
        return self.t - self.prev_t

    def update_poses(self, t: float, new_poses: Dict[Entity, np.ndarray]) -> None:
        """
        Update poses of all entities.

        The poses dictionary will be replaced with the new_poses and previous poses
        will be updated for those entities in new_poses. For an entity in new_poses
        which is not in the current poses (i.e. a new entity) the previous pose will
        be estimated using trajectory extrapolation (to correctly calculate the
        initial velocity).
        """
        self.t = t

        prev_poses = {}
        for e in new_poses:
            if e in self.poses:
                prev_poses[e] = self.poses[e]
            elif self.prev_t is not None:
                prev_poses[e] = e.trajectory.position_at_t(
                    self.prev_t, extrapolate=True
                )
        self.prev_poses = prev_poses
        self.poses = new_poses
        if self.prev_t is not None:
            self.update_statistics()
        for entity, pose in self.poses.items():
            self._recorded_poses[entity].append((self.t, pose))

    def update_statistics(self) -> None:
        """Update entity velocities and distance travelled."""
        self.velocities = {}
        for entity in self.poses:
            self.velocities[entity] = (
                self.poses[entity] - self.prev_poses[entity]
            ) / self.dt
            self.distances[entity] += np.linalg.norm(
                (self.poses[entity] - self.prev_poses[entity])[:3]
            )

    def update_actions(self) -> None:
        """Update state actions."""
        unapplied: List[ScenarioAction] = []
        for act in self.unapplied_actions:
            if act.trigger_condition(self):
                self.apply_action(act)
                self.action_apply_times[act] = self.t
            else:
                unapplied.append(act)
        self.unapplied_actions = unapplied

    def apply_action(self, action: ScenarioAction) -> None:
        """Apply an action to the state."""
        entity = self.scenario.entity_by_name(action.entity_ref)
        if entity is None:
            warnings.warn(
                f"No entity with name {entity.ref} was found for action "
                f"{action.__class__.__name__}."
            )
        else:
            action.apply(self, entity)

    def update_callbacks(self) -> None:
        """Update all state callbacks."""
        for m in self.state_callbacks:
            m(self)

    def check_terminal(self) -> bool:
        """Check if the state is terminal."""
        return any(cond(self) for cond in self.terminal_conditions)

    def recorded_poses(
        self,
        entity: Optional[Entity] = None,
    ) -> Union[np.ndarray, Dict[Entity, np.ndarray]]:
        """Get recorded poses for each or a given entity."""
        if entity is not None:
            poses = self._recorded_poses.get(entity, None)
            if not poses:
                return np.empty((0, 7))
            ts, poses = map(np.array, zip(*poses))
            return np.concatenate([ts[:, None], poses], axis=1)
        data: Dict[Entity, np.ndarray] = {}
        for ent, poses in self._recorded_poses.items():
            if not poses:
                data[ent] = np.empty((0, 7))
            else:
                ts, poses = map(np.array, zip(*poses))
                data[ent] = np.concatenate([ts[:, None], poses], axis=1)
        return data

    def get_entity_data(
        self, entity: Entity
    ) -> Tuple[float, float, np.ndarray, np.ndarray, float, np.ndarray, Any]:
        """Get state data for a specific entity."""
        return (
            self.t,
            self.next_t,
            self.poses.get(entity, None),
            self.velocities.get(entity, None),
            self.distances.get(entity, None),
            self.recorded_poses(entity=entity),
            self.entity_state.get(entity, None),
        )

    def collisions(self) -> Dict[Entity, List[Entity]]:
        """Return collisions between entities at the current time."""
        if self._collisions is None:
            self._collisions = detect_collisions(self.poses)
        return self._collisions

    def get_callback(
        self, Callback: Type[StateCallback]
    ) -> Optional[StateCallback]:
        """Get a particular type of callback."""
        if Callback not in self._callbacks:
            for callback in self.state_callbacks:
                if isinstance(callback, Callback):
                    self._callbacks[Callback] = callback
        return self._callbacks.get(Callback)

    def get_entity_box_points(self, e: Entity) -> np.ndarray:
        """Get the coordinates of the bounding box of an entity."""
        return e.get_bounding_box_points(self.poses[e])

    def get_entity_box_geom(self, e: Entity) -> Polygon:
        """Get the geometry of the bounding box of an entity."""
        return e.get_bounding_box_geom(self.poses[e])

    def get_road_info_at_entity(
        self, e: Entity
    ) -> Tuple[List[str], List[RoadObject]]:
        """Return the road network information at the entities location."""
        if not self.scenario.road_network:
            return [], []
        return self.scenario.road_network.get_geometries_at_point(
            *self.poses[e][:2]
        )

    def get_entities_in_area(
        self, area: Union[MultiPolygon, Polygon]
    ) -> List[Entity]:
        """
        Return all entities who's center point is within an area.

        Parameters
        ----------
        area : Union[MultiPolygon, Polygon]
            A shapely geometry covering the chosen area.

        """
        pos = np.array([pose[:2] for pose in self.poses.values()])
        in_area = contains(area, pos[:, 0], pos[:, 1])
        return [e for e, b in zip(self.poses, in_area) if b]

    def get_entities_in_radius(self, x: float, y: float, r: float) -> List[Entity]:
        """
        Get entities with center point within a circle.

        Parameters
        ----------
        x : float
            The x-coordinate of the center.

        y : float
            The x-coordinate of the center.

        r : float
            The radius of the center.

        """
        return self.get_entities_in_area(Point(x, y).buffer(r))

    def to_scenario(self, name: Optional[str] = None) -> Scenario:
        """Create a scenario from the historical data in the state."""
        if name is None:
            name = (
                f"Simulation of {self.scenario.name}"
                if self.scenario.name is None
                else None
            )
        entities = []
        for entity, poses in self.recorded_poses().items():
            new_entity = deepcopy(entity)
            if is_stationary(poses):
                poses = poses[None, 0]
            new_entity.trajectory = Trajectory(poses)
            entities.append(new_entity)
        return Scenario(
            entities,
            name=name,
            road_network=self.scenario.road_network,
            actions=self.scenario.actions,
        )


TERMINAL_CONDITIONS = {
    "max_length": lambda s: s.t + s.dt > s.scenario.length,
    "collision": lambda s: any(len(l) > 0 for l in s.collisions().values()),
    "ego_collision": lambda s: len(s.collisions()[s.scenario.entities[0]]) > 0,
    "ego_off_road": lambda s: not (
        s.scenario.road_network.driveable_surface.contains(
            Point(*s.poses[s.scenario.entities[0]][:2])
        )
        if s.scenario.entities[0] in s.poses
        else False
    ),
}



from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np

from scenario_gym.action import Action, TeleportAction, VehicleAction
from scenario_gym.entity import Entity
from scenario_gym.state import State
from scenario_gym.utils import ArrayLike


class Controller(ABC):
    """
    Base class for the controller. Takes the agent's action and returns the pose.

    When implementing a controller the _step method should return the
    new pose for the entity (as an np.ndarray). It may modify the state
    in other ways but the controller's step method will update the
    entities pose. This is to avoid errors with immutable arrays.
    """

    def __init__(self, entity: Entity):
        """Construct the controller from the entity."""
        self.entity = entity

    def reset(self, state: State) -> None:
        """Reset the controller parameters."""
        self._reset(state)

    def step(self, state: State, action: Action) -> ArrayLike:
        """Return the agent's next pose from the action."""
        return self._step(state, action)

    @abstractmethod
    def _reset(self, state: State) -> None:
        """Reset the controller parameters."""
        pass

    @abstractmethod
    def _step(self, state: State, action: Action) -> ArrayLike:
        """Return the agent's next pose from the action."""
        pass


class ReplayTrajectoryController(Controller):
    """A controller to replay preset trajectories."""

    def _reset(self, state: State) -> None:
        """Reset the controller parameters."""
        pass

    def _step(self, state: State, action: TeleportAction) -> ArrayLike:
        """Return the agent's next pose from the action."""
        return action.pose


class VehicleController(Controller):
    """
    A vehicle controller using a simple physical model.

    Allows acceleration and steering in a given range.
    """

    def __init__(
        self,
        entity: Entity,
        max_steer: float = 0.7,
        max_accel: float = 5.0,
        max_speed: Optional[float] = None,
        allow_reverse: bool = False,
    ):
        """
        Construct the controller from the entity.

        Parameters
        ----------
        entity : Entity
            The entity for the controller.

        max_steer : float
            The max allowed (absolute) steering angle.

        max_accel : float
            The max allowed (absolute) acceleration.

        max_speed : Optional[float]
            If given then the entity is limited to this max speed.

        allow_reverse : bool
            Allow the vehicle to move backwards. If False then the
            vehicle's speed is forced >= 0.

        """
        super().__init__(entity)
        self.max_steer = max_steer
        self.max_accel = max_accel
        self.allow_reverse = allow_reverse
        self.max_speed = max_speed

    def _reset(self, state: State) -> None:
        """Reset the controller parameters."""
        self.speed = np.linalg.norm(state.velocities[self.entity][:2])
        self.l = self.entity.catalog_entry.bounding_box.length

    def _step(
        self, state: State, action: Union[VehicleAction, np.ndarray]
    ) -> ArrayLike:
        """
        Return the agent's next pose from the action.

        Updates the heading based on the steering angle. Then calculates
        the new speed to return the new velocity.
        """
        if isinstance(action, VehicleAction):
            accel, steer = action.acceleration, action.steering
        else:
            accel, steer = action

        accel = np.clip(accel, -self.max_accel, self.max_accel)
        steer = np.clip(steer, -self.max_steer, self.max_steer)

        pose = state.poses[self.entity].copy()
        dt = state.next_t - state.t
        h = pose[3]

        dx = self.speed * np.cos(h)
        dy = self.speed * np.sin(h)
        dh = self.speed * np.tan(steer) / self.l

        pose[[0, 1]] += np.array([dx, dy]) * dt
        pose[3] += dh * dt

        speed = self.speed + accel * dt
        if not self.allow_reverse:
            speed = np.maximum(0.0, speed)
        if self.max_speed is not None:
            speed = np.minimum(self.max_speed, speed)
        self.speed = speed

        return pose


class PIDController(VehicleController):
    """
    A PID controller for scenario gym agents.

    Selects acceleration and steering to get to a given waypoint. These
    are computed using a very simple PID controller for the steering
    and acceleration. The acceleration error and steering errors are based
    on the lateral and longitudinal error from the target in the vehicles
    local frame.
    """

    def __init__(
        self,
        entity: Entity,
        steer_Kp: float = 0.03054,
        steer_Kd: float = 1.5709,
        accel_Kp: float = 0.3753,
        accel_Kd: float = 1.8970,
        accel_Ki: float = 0.0204,
        **kwargs,
    ):
        """
        Construct the controller from the entity.

        entity : Entity
            The entity for the controller.

        steer_Kp : float
            The steering proportionality parameter.

        steer_Kd : float
            The steering derivative parameter.

        accel_Kp : float
            The acceleration proportionality parameter.

        accel_Kd : float
            The acceleration derivative parameter.

        accel_Ki : float
            The acceleration integral parameter.

        kwargs:
            Keyword arguments for the underlying vehicle model.
        """
        super(self.__class__, self).__init__(
            entity,
            **kwargs,
        )
        self.steer_Kp = steer_Kp
        self.steer_Kd = steer_Kd
        self.accel_Kp = accel_Kp
        self.accel_Ki = accel_Ki
        self.accel_Kd = accel_Kd

    def _reset(self, state: State) -> None:
        """Reset the controller parameters."""
        self.e_lon_prev = 0.0
        self.e_lon_int = 0.0
        self.e_lat_prev = 0.0
        super(self.__class__, self)._reset(state)

    def _step(self, state: State, action: TeleportAction) -> ArrayLike:
        """
        Return the agent's next pose from the action.

        Calculates the longitudinal and lateral error to use as error
        values for the acceleration and steering. Then the parameters
        are applied to produce vehicle action values for the vehicle
        controller.
        """
        # current and target positions
        target = action.pose[:2]
        pose = state.poses[self.entity].copy()
        cur, h = pose[:2], pose[3]
        speed = self.speed

        # error and derivatives
        e = target - cur  # (2,)
        R = np.array(
            [
                [np.cos(h), np.sin(h)],
                [-np.sin(h), np.cos(h)],
            ]
        )
        e_lon, e_lat = R.dot(e)

        # steering
        if speed > 5.0 and speed <= 15:
            gain_adj = 1.0 - 0.9 * ((speed - 5.0)) / 10.0
        elif speed > 15:
            gain_adj = 0.1
        else:
            gain_adj = 1.0

        e_lat_D = (e_lat - self.e_lat_prev) / state.dt
        steer_Kp = self.steer_Kp * gain_adj
        steer_Kd = self.steer_Kd * gain_adj
        steer = steer_Kp * e_lat + steer_Kd * e_lat_D

        # acceleration
        e_lon_D = (e_lon - self.e_lon_prev) / state.dt
        e_lon_I = self.e_lon_int + e_lon * state.dt
        if abs(e_lon) > 0.1:
            accel = (
                self.accel_Kp * e_lon
                + self.accel_Kd * e_lon_D
                + self.accel_Ki * e_lon_I
            )
        else:
            accel = 0.0

        self.e_lat_prev = e_lat
        self.e_lon_prev = e_lon
        self.e_lon_int = e_lon_I
        return super(self.__class__, self)._step(state, VehicleAction(accel, steer))



from typing import Dict, List, Optional, TypeVar

import numpy as np
from scipy.interpolate import interp1d

from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import ArrayLike

from .base import Entity

State = TypeVar("State")


class BatchReplayEntity:
    """
    A single object used to represent multiple entities.

    Will replay exact trajectories from OpenScenario files. Computation is
    vectorized for efficiency.
    """

    def __init__(
        self,
        timestep: Optional[float] = None,
        persist: bool = False,
    ):
        """Init the batch entity with no assigned entities."""
        self.entities: List[Entity] = []
        self.trajectories: List[Trajectory] = []
        self.persist = persist
        self.timestep = timestep
        self.max_t = 0.0

    def step(self, state: State) -> Dict[Entity, ArrayLike]:
        """
        Take a single step in the gym.

        Returns the pose of each entity at the next timestamp. If enduring entities
        is set to True then only entities present at the current time will be
        returned.
        """
        t = state.next_t
        new_poses = {}
        if len(self.entities) > 0:
            pos = self.fn(t)  # (m, num_ents)
            for e, p in zip(self.entities, pos):
                if (
                    self.persist
                    or e.is_static()
                    or (t >= e.trajectory.min_t and t <= e.trajectory.max_t)
                ):
                    new_poses[e] = p
        return new_poses

    def add_entities(
        self,
        entities: List[Entity],
        trajs: List[Trajectory],
    ) -> None:
        """
        Add entities that are to be batched together.

        This will reset the entities in the scenario so all entities
        must be passed at once.

        Parameters
        ----------
        entities : List[Entity]
            The entities to be used.

        trajs : List[Trajectory]
            The trajectory for each entity.

        """
        self.entities.clear()
        self.trajectories.clear()
        self.max_t = 0.0
        if entities:
            self.entities.extend(entities)
            self.trajectories.extend(trajs)

            num_ents = len(self.entities)
            datas = []
            for t in self.trajectories:
                d = np.nan_to_num(t.data)
                if d.shape[0] == 1:
                    d = np.repeat(d, 2, axis=0)
                    d[-1, 0] += 1e-1  # to prevent nan
                datas.append(d)

            m = datas[0].shape[1] - 1
            ts = np.array(
                sorted(list(set([t for d in datas for t in d[:, 0]])))
            )  # (N,)
            self.max_t = ts[-1]

            interpd = []
            for d in datas:
                x = interp1d(
                    d[:, 0],
                    d[:, 1:].T,
                    bounds_error=False,
                    fill_value=(d[0, 1:], d[-1, 1:]),
                )(
                    ts
                ).T  # (N, m)
                interpd.append(x)

            X = np.concatenate(interpd, axis=1)  # (N, num_ents * m)
            if self.timestep:
                all_ts = np.arange(0.0, self.max_t, self.timestep)
                all_Xs = interp1d(
                    ts,
                    X.T,
                    bounds_error=False,
                    fill_value=(X[0], X[-1]),
                )(all_ts).T
                self.fn = lambda t: all_Xs[np.abs(all_ts - t).argmin()].reshape(
                    num_ents, m
                )
            else:
                interp = interp1d(
                    ts,
                    X.T,
                    bounds_error=False,
                    fill_value=(X[0], X[-1]),
                )
                self.fn = lambda t: interp(t).reshape(num_ents, m)



import numpy as np
import pytest as pt

from scenario_gym.catalog_entry import BoundingBox, CatalogEntry
from scenario_gym.entity import Entity
from scenario_gym.scenario import Scenario
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.state import detect_collisions
from scenario_gym.trajectory import Trajectory


@pt.fixture
def collision_scenario():
    """Create a scenario with two entities that collide."""
    box = BoundingBox(2.0, 5.0, 0.0, 0.0)
    ce = CatalogEntry("car", "car", "car", "car", box, {}, [])
    ego = Entity(ce, ref="ego")
    hazard = Entity(ce, ref="entity_1")

    ego.trajectory = Trajectory(
        np.array(
            [
                [0.0, 0, 0],
                [10, 20, 0],
            ]
        ),
        fields=["t", "x", "y"],
    )
    hazard.trajectory = Trajectory(
        np.array(
            [
                [0.0, 40, 0],
                [10, 20, 0],
            ]
        ),
        fields=["t", "x", "y"],
    )

    scenario = Scenario([ego, hazard])
    return scenario, ego, hazard


def test_detect_collisions(collision_scenario):
    """Test collisions are correctly detected."""
    s, ego, hazard = collision_scenario
    gym = ScenarioGym()
    gym.set_scenario(s)

    collisions = detect_collisions(
        {ego: gym.state.poses[ego]},
        others={hazard: gym.state.poses[hazard]},
    )
    assert not collisions[ego], "No collision at start of scenario"

    gym.rollout()

    collisions = detect_collisions(
        {ego: gym.state.poses[ego]},
        others={hazard: gym.state.poses[hazard]},
    )
    assert collisions[ego], "Collision at end of scenario not found."



import math
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np
from shapely.geometry import MultiPolygon
from shapely.ops import unary_union
from shapely.prepared import prep
from shapely.vectorized import contains

from scenario_gym.entity import Entity
from scenario_gym.observation import SingleEntityObservation
from scenario_gym.road_network import RoadNetwork
from scenario_gym.state import State
from scenario_gym.utils import ArrayLike, NDArray

from .base import Sensor


@dataclass
class MapObservation(SingleEntityObservation):
    """Observation with a raster map."""

    map: np.ndarray


class RasterizedMapSensor(Sensor):
    """
    Returns a rasterized semantic map as a 2d grid of vectors.

    Additional custom layers may be implemented by subclassing this sensor
    and implementing a perpare and a get method for the new layer.

    The prepare method should be called `_prepare_{}_layer` where {} is replaced
    with the string layer name. This method is called once when the road network
    is first seen and can be used to prepare any required data e.g. using prep on
    any shapely objects to improve performance.

    The get method should be called `_{}_layer` where {} is replaced with the
    string layer name. This method is called each step with the state and the map
    coordiantes to return the value of the map at each coordinate.
    """

    _all_layers: List[str] = [
        "entity",
        "driveable_surface",
        "road",
        "intersection",
        "lane",
        "walkable_surface",
        "pavement",
        "crossing",
    ]

    def __init__(
        self,
        entity: Entity,
        layers: Optional[List[str]] = None,
        height: float = 20.0,
        width: float = 20.0,
        freq: Optional[float] = 1.0,
        n: Optional[int] = None,
        channels_first: bool = False,
    ):
        """
        Init the sensor.

        Parameters
        ----------
        entity : Entity
            The entity.

        layers : Optional[List[str]]
            The layers to be observed as string names. The available layers can be
            found in `RasterizedMapSensor._all_layers`. The order in which these
            are returned will be used for the output array.

        height : float
            The length of the box around the entity in the y-direction in the
            local frame.

        width : float
            The length of the box around the entity in the x-direction in the
            local frame.

        freq : int
            The frequency of sampling points for the rasterized map. Only one of
            `freq` and `n` should be passed.

        n : Optional[int]
            The number of sampling points for the rasterized map. Only one of
            `freq` and `n` should be passed.

        channels_first : bool
            If given returns (C, W, H) rather than (W, H, C)

        """
        super().__init__(entity)
        self.layers = (
            layers if layers is not None else ["entity", "driveable_surface"]
        )
        self.check_layers()

        self.height = height
        self.width = width
        self.channels_first = channels_first
        if n is None:
            assert freq is not None, "At least one of n and freq must be provided."
            self.nw, self.nh = int(freq * width), int(freq * height)
        else:
            self.nw = self.nh = n

        self.X = np.array(
            np.meshgrid(
                np.linspace(-self.width / 2, self.width / 2, self.nw),
                np.linspace(-self.height / 2, self.height / 2, self.nh),
            )
        ).transpose(1, 2, 0)

    def check_layers(self) -> None:
        """Check that all layers are implemented correctly."""
        for layer in self.layers:
            try:
                getattr(self, f"_{layer}_layer")
                getattr(self, f"_prepare_{layer}_layer")
            except AttributeError:
                raise NotImplementedError(
                    f"Layer {layer} does not have a get and/or a prepare method."
                )

    def _reset(self, state: State) -> MapObservation:
        """Reset the sensor at the start of the scenario."""
        self._road_network: Optional[RoadNetwork] = None
        return self._step(state)

    def _step(self, state: State) -> MapObservation:
        """Return the rasterized map around the entity."""
        if self._road_network is None:
            self._prepare_layers(state)

        pose = state.poses[self.entity]
        coords = self._get_coords(pose).reshape(-1, 2)
        layers = [getattr(self, f"_{l}_layer")(state, coords) for l in self.layers]
        obs_map = np.array(layers).reshape(len(layers), self.nw, self.nw)
        return MapObservation(
            self.entity,
            *state.get_entity_data(self.entity),
            obs_map if self.channels_first else obs_map.transpose(1, 2, 0),
        )

    @property
    def output_shape(self) -> Tuple[int, int, int]:
        """Return the output shape of the rasterized map."""
        if self.channels_first:
            return (len(self.layers), self.nw, self.nh)
        return (self.nw, self.nh, len(self.layers))

    def _get_coords(self, pose: ArrayLike) -> NDArray:
        """Get the coordinates at which the map should be constructed."""
        X = self.X  # (nw, nh, 2)

        xy, theta = pose[[0, 1]], pose[3] + math.pi / 2
        R = np.array(
            [
                [np.cos(theta), -np.sin(theta)],
                [np.sin(theta), np.cos(theta)],
            ]
        )
        return (X @ R.T) + xy[None, None, :]

    def _prepare_layers(self, state: State) -> None:
        """Get all data needed to compute the map at future timesteps."""
        self._road_network = state.scenario.road_network
        for layer in self.layers:
            getattr(self, f"_prepare_{layer}_layer")(state)

    def _prepare_entity_layer(self, state: State) -> None:
        """Prepare the entity layer."""
        pass

    def _entity_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """
        Check which points are occupied by the bounding box of an entity.

        Note: this includes the sensor's own entity.
        """
        entities = prep(
            MultiPolygon(
                [e.get_bounding_box_geom(pose) for e, pose in state.poses.items()]
            )
        )
        return contains(entities, coords[:, 0], coords[:, 1])

    def _prepare_driveable_surface_layer(self, state: State) -> None:
        """Prepare the driveable surface layer."""
        self._driveable_surface = prep(self._road_network.driveable_surface)

    def _driveable_surface_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which of the given points lie in the driveable surface."""
        return contains(self._driveable_surface, coords[:, 0], coords[:, 1])

    def _prepare_road_layer(self, state: State) -> None:
        """Prepare the road layer."""
        self._roads = prep(
            unary_union(
                [r.boundary for r in self._road_network.roads],
            )
        )

    def _road_layer(self, state: State, coords: ArrayLike) -> ArrayLike:
        """Check which points lie in a road."""
        return contains(self._roads, coords[:, 0], coords[:, 1])

    def _prepare_intersection_layer(self, state: State) -> None:
        """Prepare the intersection layer."""
        self._intersections = prep(
            unary_union(
                [i.boundary for i in self._road_network.intersections],
            )
        )

    def _intersection_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which points lie in an intersection."""
        return contains(self._intersections, coords[:, 0], coords[:, 1])

    def _prepare_lane_layer(self, state: State) -> None:
        """Prepare the lane layer."""
        self._lanes = prep(
            unary_union(
                [l.boundary for r in self._road_network.roads for l in r.lanes],
            )
        )

    def _lane_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which points lie in a lane."""
        return contains(self._lanes, coords[:, 0], coords[:, 1])

    def _prepare_walkable_surface_layer(self, state: State) -> None:
        """Prepare the walkable surface layer."""
        self._walkable_surface = prep(self._road_network.walkable_surface)

    def _walkable_surface_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which points lie in a walkable surface."""
        return contains(self._walkable_surface, coords[:, 0], coords[:, 1])

    def _prepare_pavement_layer(self, state: State) -> None:
        """Prepare the pavement layer."""
        self._pavements = prep(
            unary_union([p.boundary for p in self._road_network.pavements])
        )

    def _pavement_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which points lie in a pavement."""
        return contains(self._pavements, coords[:, 0], coords[:, 1])

    def _prepare_crossing_layer(self, state: State) -> None:
        """Prepare the crossing layer."""
        self._crossings = prep(
            unary_union([c.boundary for c in self._road_network.crossings])
        )

    def _crossing_layer(self, state: State, coords: ArrayLike) -> NDArray:
        """Check which points lie in a pedestrian crossing."""
        return contains(self._crossings, coords[:, 0], coords[:, 1])



import numpy as np
import pytest as pt

from scenario_gym.trajectory import Trajectory, _resolve_heading


def test_trajectory():
    """Test that trajectories are constructed correctly."""
    data = np.repeat(np.arange(10, dtype=np.float32)[:, None], 4, axis=1)  # (10, 4)
    traj = Trajectory(data, fields=["t", "x", "y", "h"])
    assert traj.h is not None
    assert traj.max_t == 9, f"Incorrect max time: {traj.max_t}."
    assert np.allclose(
        traj.arclength, 9 * np.sqrt(2)
    ), f"Incorrect arclength: {traj.arclength}."


def test_invalid_create():
    """Test that we cannot create without txy or with an invalid shape."""
    with pt.raises(ValueError):
        Trajectory(np.empty((3, 2)), fields=["x", "y"])
        Trajectory(np.empty((3, 2)), fields=["t", "y"])
        Trajectory(np.empty((3, 2)), fields=["t", "x"])
        Trajectory(np.empty((2)))
        Trajectory(np.empty((2, 2, 2)))
        Trajectory(np.empty((2, 4)), fields=["t", "x", "y"])
        Trajectory(np.empty((2, 2)))
        Trajectory(np.array([[0, np.nan, 0]]), fields=["t", "x", "y"])
        Trajectory(np.array([[0, 0.0, np.nan, 0]]), fields=["t", "h", "x", "y"])


def test_unordered_create():
    """Test creating a trajectory with fields in a different order."""
    data = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 1.0],
            [0.0, 1.0, 0.0, 2.0],
            [1.0, 2.0, 0.0, 3.0],
        ]
    )
    traj = Trajectory(data, fields=["y", "x", "h", "t"])
    assert np.allclose(traj.t, [0.0, 1.0, 2.0, 3.0]), "t not set correctly."
    assert np.allclose(traj.x, [0.0, 0.0, 1.0, 2.0]), "x not set correctly."
    assert np.allclose(traj.y, [0.0, 0.0, 0.0, 1.0]), "y not set correctly."
    assert np.allclose(traj.h, [0.0, 1.0, 0.0, 0.0]), "h not set correctly."


def test_create_with_duplicate_nan():
    """Test creating with duplicate rows containing nan values."""
    traj = Trajectory(
        np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, np.nan],
                [1.0, 0.0, 0.0, np.nan],
                [2.0, 0.0, 0.0, 0.0],
            ]
        ),
        fields=["t", "x", "y", "h"],
    )
    assert traj.data.shape[0] == 3, "Duplicate rows should be removed."


def test_filled_headings():
    """Test that headings are estimated when not provided."""
    data = np.array(
        [
            [0, 0, 0],
            [1, 1, 1],
            [2, 2, 2],
        ]
    )
    traj = Trajectory(data, fields=["t", "x", "y"])
    true_h = np.ones(3) * np.pi / 4
    assert np.allclose(traj.h, true_h), "Headings should be estimated from xy."


def test_filled_zrp():
    """Test that angles are estimated when not provided."""
    data = np.array(
        [
            [0, 0, 0],
            [1, 1, 1],
            [2, 2, 2],
        ]
    )
    traj = Trajectory(data, fields=["t", "x", "y"])
    assert np.allclose(traj.z, np.zeros(3)), "z not filled correctly."
    assert np.allclose(traj.z, np.zeros(3)), "r not filled correctly."
    assert np.allclose(traj.z, np.zeros(3)), "p not filled correctly."


def test_single_cp():
    """Test loading a trajectory with 1 control point."""
    data = np.array(
        [
            [
                0.0,
                1.0,
                1.0,
            ]
        ]
    )
    traj = Trajectory(data, fields=["t", "x", "y"])
    assert np.allclose(traj.h, np.zeros(1)), "Should be assigned 0 heading."
    assert np.allclose(
        traj.position_at_t(10.0)[:2], np.ones(2)
    ), "Should stay fixed."
    assert np.allclose(traj.max_t, 0.0), "Max t should be 0."


def test_resolve_heading():
    """Test that resolve headings correctly reduces the gap between headings."""
    data = np.random.randn(100)
    resolved = _resolve_heading(data)

    delta = (data - resolved) % (2 * np.pi)
    delta = np.where(delta >= np.pi, delta - 2 * np.pi, delta)
    assert np.allclose(
        delta, np.zeros(data.shape[0])
    ), "Difference should be 0 modulo 2pi."
    assert (
        np.abs(np.diff(delta)).sum() <= np.abs(np.diff(data)).sum()
    ), "Stepwise distance should be reduced."


def test_immutable_traj():
    """Test that trajectories are copied correctly."""
    data = np.repeat(np.arange(10, dtype=np.float32)[:, None], 4, axis=1)  # (10, 4)
    traj = Trajectory(data, fields=["t", "x", "y", "h"])

    with pt.raises(ValueError):
        traj.data[-1][0] = 11


def test_copy_traj():
    """Test copying a trajectory."""
    data = np.repeat(np.arange(10, dtype=np.float32)[:, None], 4, axis=1)  # (10, 4)
    traj = Trajectory(data, fields=["t", "x", "y", "h"])

    traj_new = traj.copy()
    assert id(traj) != id(traj_new), "Should have different memory."
    assert id(traj.data) != id(traj_new.data), "Should have different memory."
    assert np.allclose(traj.data, traj_new.data), "Should have equal data."


def test_position_at_t():
    """Test the position at t method."""
    data = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    traj = Trajectory(data, fields=["t", "x", "y"])

    # test basic interpolation
    assert np.allclose(
        traj.position_at_t(0.5, extrapolate=True)[:2], [0.5, 0.5]
    ), "Should be 0.5."
    assert np.allclose(
        traj.position_at_t(1.5, extrapolate=True)[:2], [1.5, 1.5]
    ), "Should be 1.5."

    # test position with int
    traj.position_at_t(1)

    # test simple extrapolation
    assert np.allclose(
        traj.position_at_t(2.5, extrapolate=True)[:2], [2.5, 2.5]
    ), "Should be 2.5."
    assert np.allclose(
        traj.position_at_t(-1.0, extrapolate=True)[:2], [-1.0, -1.0]
    ), "Should be -1."
    assert traj.position_at_t(-1.0, extrapolate=False) is None, "Should be None."
    assert traj.position_at_t(3.0, extrapolate=False) is None, "Should be None."
    assert np.allclose(
        traj.position_at_t(-1.0, extrapolate=(False, True))[:2], [0.0, 0.0]
    ), "Should not extrapolate backward."
    assert np.allclose(
        traj.position_at_t(-1.0, extrapolate=(True, True))[:2], [-1.0, -1.0]
    ), "Should extrapolate backward."
    assert np.allclose(
        traj.position_at_t(3.0, extrapolate=(True, False))[:2], [2.0, 2.0]
    ), "Should not extrapolate forward."
    assert np.allclose(
        traj.position_at_t(3.0, extrapolate=(True, True))[:2], [3.0, 3.0]
    ), "Should extrapolate forward."

    # test broadcasting
    assert np.allclose(
        traj.position_at_t(data[:, 0], extrapolate=True)[:, :2], data[:, 1:]
    ), "Incorrect broadcasting."
    x = np.array([-1.0, 3.0])
    assert np.allclose(
        traj.position_at_t(x, extrapolate=True)[:, :2],
        np.array([[-1.0, -1.0], [3.0, 3.0]]),
    ), "Incorrect extrapolation."
    assert np.allclose(
        traj.position_at_t(x, extrapolate=False)[:, :2],
        np.array([[0.0, 0.0], [2.0, 2.0]]),
    ), "Incorrect extrapolation."
    assert np.allclose(
        traj.position_at_t(x, extrapolate=(False, True))[:, :2],
        np.array([[0, 0], [3.0, 3.0]]),
    ), "Incorrect extrapolation."
    assert np.allclose(
        traj.position_at_t(x, extrapolate=(True, False))[:, :2],
        np.array([[-1.0, -1.0], [2.0, 2.0]]),
    ), "Incorrect extrapolation."
    assert np.allclose(
        traj.position_at_t(x, extrapolate=(True, True))[:, :2],
        np.array([[-1.0, -1.0], [3.0, 3.0]]),
    ), "Incorrect extrapolation."


def test_velocity_at_t():
    """Test the position at t method."""
    data = np.array([[0, 0, 0], [1, 0, 1], [2, 0, 2]])
    traj = Trajectory(data, fields=["t", "x", "y"])

    # test basic interpolation
    assert np.allclose(traj.velocity_at_t(0.5)[:2], [0, 1]), "Should be 0, 1."
    assert np.allclose(
        traj.velocity_at_t([0.5, 2.5])[:, :2], [[0, 1], [0, 0]]
    ), "Incorrect broadcasting.."


def test_subsample():
    """Test subsampling a trajectory."""
    traj = Trajectory(
        np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 0.0],
                [2.0, 2.0, 0.0],
                [3.0, 3.0, 0.0],
                [4.0, 4.0, 0.0],
            ]
        ),
        fields=["t", "x", "y"],
    )
    subsample_t = traj.subsample(points_per_t=0.5)
    assert all(
        (
            subsample_t.min_t == 0.0,
            subsample_t.max_t == 4.0,
            subsample_t.arclength == 4.0,
        )
    ), "Incorrect trajectory produced."

    subsample_s = traj.subsample(points_per_s=0.5)
    assert all(
        (
            subsample_s.min_t == 0.0,
            subsample_s.max_t == 4.0,
            subsample_s.arclength == 4.0,
        )
    ), "Incorrect trajectory produced."


def test_curvature_subsample():
    """Test subsampling a trajectory."""
    base_traj = np.array(
        [[0.0, 0.0, 0.0], [1.0, 1.0, 0.0], [1.9, 1.9, 0.0]],
    )
    mid_y = np.linspace(0, 2, 20)
    mid_x = 2 + np.sqrt(1 - (mid_y - 1) ** 2)
    middle = np.array([np.linspace(2.0, 4.0, 20), mid_x, mid_y]).T
    end = np.array(
        [[4.1, 1.9, 2.0], [6.0, 0.0, 2.0]],
    )
    traj = Trajectory(np.vstack([base_traj, middle, end]), fields=["t", "x", "y"])
    subsampled = traj.curvature_subsample(points_per_s=5)
    assert all(
        (
            subsampled.min_t == 0.0,
            subsampled.max_t == 6.0,
        )
    ), "Incorrect trajectory produced."
    assert (subsampled.x >= 2).sum() > 0.5 * subsampled.x.shape[
        0
    ], "More points should be in the final part of the trajectory."



from itertools import chain
from typing import Dict, List, Optional

import numpy as np

from scenario_gym.entity import Entity
from scenario_gym.utils import detect_geom_collisions


def detect_collisions(
    entities: Dict[Entity, np.ndarray],
    others: Optional[Dict[Entity, np.ndarray]] = None,
) -> Dict[Entity, List[Entity]]:
    """
    Return collisions between entities.

    Checks the bounding boxes of all entities at the current time
    and return any that overlap. Returns a dict of entities in
    the scenario with a list of all other entities that they are
    colliding with.

    Parameters
    ----------
    entities : List[Entity]
        The entities.

    others : Optional[List[Entity]]
        Additional entities to include when checking for collisions
        with each geometry in geoms.

    """
    geom_to_ent = {}
    for e, pose in (
        entities.items()
        if others is None
        else chain(entities.items(), others.items())
    ):
        g = e.get_bounding_box_geom(pose)
        geom_to_ent[g] = e
        geom_to_ent[e] = g

    geoms = [geom_to_ent[e] for e in entities]
    other_geoms = [geom_to_ent[e] for e in others] if others is not None else None

    collisions = detect_geom_collisions(geoms, others=other_geoms)
    return {
        e: [geom_to_ent[g_prime] for g_prime in collisions[g]]
        for e, g in zip(entities, geoms)
    }



from copy import copy

import numpy as np
import pytest as pt

from scenario_gym import Trajectory
from scenario_gym.entity.base import Entity, StaticEntity
from scenario_gym.entity.batch import BatchReplayEntity
from scenario_gym.xosc_interface import import_scenario


@pt.fixture(scope="module")
def example_catalog_entry(all_scenarios):
    """Get an example catalog entry for use in entity tests."""
    scenario = import_scenario(
        all_scenarios["3fee6507-fd24-432f-b781-ca5676c834ef"]
    )
    return scenario.entities[0].catalog_entry


def test_copy_entity(example_catalog_entry):
    """Test deepcopying an entity."""
    e = Entity(
        example_catalog_entry,
        trajectory=Trajectory(
            np.array([[0.0, 1.0, 2.0]]),
            fields=["t", "x", "y"],
        ),
        ref="example",
    )
    e_new = copy(e)
    assert id(e_new.catalog_entry) == id(
        e.catalog_entry
    ), "Catalog entry should be the same."
    assert (id(e_new.trajectory) != id(e.trajectory)) and np.allclose(
        e_new.trajectory.data, e.trajectory.data
    ), "Trajectories should be the equal but different memory."


def test_batch_entity(example_catalog_entry):
    """Test deepcopying an entity."""
    e1 = Entity(
        example_catalog_entry,
        trajectory=Trajectory(
            np.array([[0.0, 0.0, 0.0]]),
            fields=["t", "x", "y"],
        ),
        ref="example",
    )
    e2 = Entity(
        example_catalog_entry,
        trajectory=Trajectory(
            np.array([[0.0, 1.0, 0.0], [2.0, 2.0, 0.0]]),
            fields=["t", "x", "y"],
        ),
        ref="example2",
    )

    class fake_state:
        """Fake state for testing."""

        next_t = 0.0

    batch = BatchReplayEntity()
    batch.add_entities([e1, e2], [e1.trajectory, e2.trajectory])

    fake_state.next_t = 1.0
    poses = batch.step(fake_state)
    assert np.allclose(poses[e1][:2], np.zeros(2)), "Entity 1 should be at origin."
    assert np.allclose(
        poses[e2][:2], np.array([1.5, 0.0])
    ), "Entity 2 should be at 1.5, 0."

    fake_state.next_t = 5.0
    poses = batch.step(fake_state)
    assert np.allclose(poses[e1][:2], np.zeros(2)), "Entity 1 should be at origin."
    assert e2 not in poses, "Entity 2 should not be returned."


def test_static_entity(example_catalog_entry):
    """Test creating a static entity."""
    e_static = StaticEntity(example_catalog_entry, ref="static_ent")
    e_static.trajectory = Trajectory(
        np.array([[0.0, 1.0, 2.0]]),
        fields=["t", "x", "y"],
    )

    with pt.raises(ValueError):
        e_static.trajectory = Trajectory(
            np.array([[0.0, 1.0, 2.0], [1.0, 2.0, 3.0]]),
            fields=["t", "x", "y"],
        )



from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.sensor.common import (
    CombinedSensor,
    EgoLocalizationSensor,
    FutureCollisionDetector,
    GlobalCollisionDetector,
    KeyboardInputDetector,
)
from scenario_gym.sensor.map import RasterizedMapSensor


def test_combined_sensor(all_scenarios):
    """Test combining a sensor."""
    s = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    gym = ScenarioGym()
    gym.load_scenario(s)
    ego = gym.state.scenario.entities[0]
    state = gym.state

    sensor = CombinedSensor(
        ego,
        EgoLocalizationSensor(ego),
        FutureCollisionDetector(ego),
        GlobalCollisionDetector(ego),
        KeyboardInputDetector(ego),
    )
    assert sensor.obs_class is None
    sensor.reset(state)
    assert sensor.obs_class is not None
    obs = sensor.step(state)

    # check attributes exist
    obs.pose
    obs.future_collision
    obs.collisions
    obs.last_keystroke


def test_map_sensor(all_scenarios):
    """Test the rasterized sensor module."""
    # load a test scenario
    s = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    gym = ScenarioGym()
    gym.load_scenario(s)
    e = gym.state.scenario.entities[0]

    # test with default layers
    sensor = RasterizedMapSensor(e, height=30, width=30, n=61)
    sensor._reset(gym.state)
    out = sensor._step(gym.state).map

    assert out.shape == (
        61,
        61,
        len(sensor.layers),
    ), f"Invalid shape: {out.shape}."
    assert out[
        ..., sensor.layers.index("driveable_surface")
    ].any(), "The ego starts on the road so the road should be in the map."
    assert out[
        30, 30, sensor.layers.index("entity")
    ], "The ego is at (0, 0) so this should be True."

    # test with all layers
    sensor = RasterizedMapSensor(
        e, layers=RasterizedMapSensor._all_layers, height=30, width=30, n=61
    )
    sensor._reset(gym.state)
    out = sensor._step(gym.state).map

    assert out.shape == (
        61,
        61,
        len(sensor.layers),
    ), f"Invalid shape: {out.shape}."
    assert out[
        ..., sensor.layers.index("driveable_surface")
    ].any(), "The ego starts on the road so the road should be in the map."
    assert out[
        30, 30, sensor.layers.index("entity")
    ], "The ego is at (0, 0) so this should be True."



from .base import RoadGeometry, RoadLike, RoadObject
from .objects import (
    Building,
    Crossing,
    Intersection,
    Lane,
    LaneType,
    Pavement,
    Road,
)
from .road_network import RoadNetwork



from dataclasses import dataclass
from typing import Optional

from lxml.etree import Element

from scenario_gym.catalog_entry import Catalog, CatalogEntry
from scenario_gym.entity import Entity, MiscObject
from scenario_gym.utils import ArgsKwargs
from scenario_gym.xosc_interface import read_catalog


@dataclass
class CustomCatalogEntry(CatalogEntry):
    """Define a custom catalog entry which has an extra property."""

    mystery_property: float

    xosc_names = ["CustomObject"]

    @classmethod
    def load_data_from_xml(
        cls,
        element: Element,
        catalog: Optional[Catalog] = None,
    ) -> ArgsKwargs:
        """Load from xml."""
        args, kwargs = super().load_data_from_xml(element, catalog=catalog)
        mystery_property = float(element.find("Mystery").attrib["value"])
        args = args + (mystery_property,)
        return args, kwargs


class CustomEntity(Entity):
    """Define a custom entity to use the custom catalog."""

    def __init__(
        self, catalog_entry: CustomCatalogEntry, ref: Optional[str] = None
    ):
        super().__init__(catalog_entry, ref=None)
        self.mystery = self.catalog_entry.mystery_property


def test_all_default_catalogs(all_catalogs):
    """Test the Scenario_Gym catalogs."""
    for catalog_file in all_catalogs.values():
        if "Scenario_Gym" in catalog_file:
            _ = read_catalog(catalog_file)


def test_custom_catalog(all_catalogs):
    """Test loading a custom catalog."""
    catalog_file = all_catalogs["Custom_Catalog/MiscCatalogs/CustomCatalog"]
    _, out = read_catalog(catalog_file, entity_types=(CustomEntity,))
    ent = out["misc_object"]
    assert ent.catalog_entry.catalog_entry == "misc_object"
    assert ent.catalog_entry.mystery_property == 100
    assert set(ent.catalog_entry.files) == set(["test.txt", "test2.txt"])


def test_misc_objects(all_catalogs):
    """Test loading a custom catalog."""
    catalog_file = all_catalogs[
        "Custom_Catalog/MiscObjectCatalogs/CustomMiscObjectCatalog"
    ]
    _, out = read_catalog(catalog_file, entity_types=(MiscObject,))
    ent = out["misc_object22"]
    assert ent.catalog_entry.catalog_entry == "misc_object22"
    assert ent.catalog_entry.mass == 1
    assert isinstance(ent, MiscObject), "Should be a MiscObject"


def test_to_xosc(all_catalogs):
    """Test loading a custom catalog."""
    catalog_file = all_catalogs[
        "Scenario_Gym/VehicleCatalogs/ScenarioGymVehicleCatalog"
    ]
    _, out = read_catalog(catalog_file)
    ent = out["car1"]
    ent_xosc = ent.catalog_entry.to_xosc()
    assert ent_xosc.name == "car1", "Should be car1."
    assert ent_xosc.vehicle_type.name == "car", "Should be car."



from typing import Any, Dict, Optional

import numpy as np
from shapely.geometry import LineString, Polygon
from shapely.validation import make_valid

from scenario_gym.utils import ArgsKwargs

from .utils import load_road_geometry_from_json, polygon_to_data


class RoadObject:
    """
    Base class for an object in the road network.

    All objects have an id attribute and implement __eq__ and
    __hash__ methods.
    """

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Create from dictionary."""
        args, kwargs = cls.load_data_from_dict(data)
        return cls(*args, **kwargs)

    @classmethod
    def load_data_from_dict(cls, data: Dict[str, Any]) -> ArgsKwargs:
        """Load raw data from dictionary."""
        return (data["Id" if "Id" in data else "id"],), {}

    def __init__(self, id: str):
        self.id = id

    def __eq__(self, other: Any) -> bool:
        """Check if another road object is the same as the current object."""
        if isinstance(other, str):
            return self.id == other
        return hasattr(other, "id") and (other.id == self.id)

    def __hash__(self) -> int:
        """Return a hash of the id."""
        return hash(self.id)

    def __repr__(self) -> str:
        """Return a repr string with the object type and its id."""
        return f"{self.__class__.__name__}(id={self.id})"

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary with id."""
        return {"id": self.id}


class RoadGeometry(RoadObject):
    """
    A geometric object in the road.

    These objects have a boundary given by a shapely polygon.

    The driveable variable indicates if vehicles should use
    the geometry. Similarly the walkable surface variable affects
    whether the geometry is included in the walkable surface. The
    impenetrable variable indicates if a geometry may not be entered
    by any entity. An instance or subclass my overwrite these variables.
    """

    driveable = True
    walkable = True
    impenetrable = False

    @classmethod
    def load_data_from_dict(cls, data: Dict[str, Any]) -> ArgsKwargs:
        """Load raw data from dictionary."""
        (obj_id,), _ = super().load_data_from_dict(data)
        boundary, _ = load_road_geometry_from_json(data)
        if "Elevation" in data and data["Elevation"] is not None:
            elevation = np.array(data["Elevation"])
        else:
            elevation = None
        return (obj_id, boundary), {"elevation": elevation}

    def __init__(
        self, id: str, boundary: Polygon, elevation: Optional[np.ndarray] = None
    ):
        super().__init__(id)

        self.boundary = self._fix_boundary(boundary)

        if elevation is not None:
            assert (
                elevation.ndim == 2 and elevation.shape[1] == 3
            ), "Invalid shape for elevation profile."
        self.elevation = elevation

    def _fix_boundary(
        self,
        boundary: Polygon,
        maxiter: int = 2000,
        tol: float = 1e-3,
    ) -> Polygon:
        """
        Fix the boundary if it is invalid.

        If the boundary is not a polygon or is not a valid geometry then tries to
        fix it with the `make_valid` function or by repeatedly buffering it with a
        small tolerance. If this fails then a ValueError is raised.
        """
        if isinstance(boundary, Polygon) and boundary.is_valid:
            return boundary
        if not boundary.is_valid:
            new = make_valid(boundary)
            if isinstance(new, Polygon) and new.is_valid:
                return new
        new = boundary.buffer(0.0)
        for _ in range(maxiter):
            new = new.buffer(tol)
            if new.is_valid and isinstance(new, Polygon):
                return new
        raise ValueError("Invalid geometry.")

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary with id and boundary."""
        data = super().to_dict()
        data["Boundary"] = polygon_to_data(self.boundary)
        data["Elevation"] = (
            self.elevation.tolist() if self.elevation is not None else None
        )
        return data


class RoadLike(RoadGeometry):
    """
    A geometry with a center line.

    Used for roads, lanes, pavements, etc.
    """

    @classmethod
    def load_data_from_dict(cls, data: Dict[str, Any]) -> ArgsKwargs:
        """Load raw data from dictionary."""
        boundary, center = load_road_geometry_from_json(data)
        if "Elevation" in data and data["Elevation"] is not None:
            elevation = np.array(data["Elevation"])
        else:
            elevation = None
        return (data["Id" if "Id" in data else "id"], boundary, center), {
            "elevation": elevation
        }

    def __init__(
        self,
        id: str,
        boundary: Polygon,
        center: LineString,
        elevation: Optional[np.ndarray] = None,
    ):
        super().__init__(id, boundary, elevation=elevation)
        self.center = center

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary with id, boundary and center."""
        data = super().to_dict()
        data["Center"] = [
            {"x": float(x), "y": float(y)} for x, y in self.center.coords
        ]
        return data



from dataclasses import dataclass

import numpy as np

from scenario_gym.observation import (
    Observation,
    SingleEntityObservation,
    combine_observations,
)
from scenario_gym.scenario_gym import ScenarioGym


def test_observation(all_scenarios) -> None:
    """Test the standard observation."""
    scenario_path = all_scenarios["a5e43fe4-646a-49ba-82ce-5f0063776566"]
    gym = ScenarioGym()
    gym.load_scenario(scenario_path)

    ego = gym.state.scenario.entities[0]
    obs = SingleEntityObservation(
        ego,
        *gym.state.get_entity_data(ego),
    )

    assert np.allclose(obs.pose, gym.state.poses[ego])
    assert np.allclose(obs.velocity, gym.state.velocities[ego])


def test_merge_observations():
    """Test merging observations from different dataclasses."""

    @dataclass
    class X:
        a: int

    @dataclass
    class Y:
        b: int

    Z = combine_observations(X, Y)

    x = X(0)
    y = Y(1)
    z = Z.from_obs(x, y)
    z2 = Z(0, 1)
    assert (z.a, z.b) == (0, 1)
    assert z == z2
    assert isinstance(z, Observation)



from __future__ import annotations

from math import inf
from types import MethodType
from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from packaging import version

from scenario_gym.action import Action
from scenario_gym.agent import Agent
from scenario_gym.controller import VehicleController
from scenario_gym.entity import Entity
from scenario_gym.observation import Observation
from scenario_gym.scenario import Scenario
from scenario_gym.scenario_gym import ScenarioGym, _create_agent
from scenario_gym.sensor.map import RasterizedMapSensor
from scenario_gym.state import TERMINAL_CONDITIONS, State

try:
    import gym
    from gym import Env
    from gym.spaces import Box, Space
except ImportError as e:
    raise ImportError(
        "gym is required for this module. Install it with `pip install gym`."
    ) from e


class ScenarioGym(ScenarioGym, Env):
    """
    An OpenAI gym compatible version of the Scenario Gym.

    Provides an explicit interface to the observations of the
    ego agent. The agent must implement the reward method of the
    Agent class.
    """

    metadata = {"render_modes": []}
    _new_reset = version.parse(gym.__version__) >= version.parse("0.22.0")

    def __init__(
        self,
        action_space: Optional[Space] = None,
        observation_space: Optional[Space] = None,
        reward_range: Tuple[float, float] = (-inf, inf),
        terminal_conditions: Optional[
            List[Union[str, Callable[[State], bool]]]
        ] = None,
        timestep: float = 0.1,
        create_agent: Optional[
            Callable[[Scenario, Entity], Optional[Agent]]
        ] = None,
        select_scenario: Optional[
            Callable[[ScenarioGym], Union[Scenario, str]]
        ] = None,
        **kwargs,
    ):
        """
        Construct the ScenarioGym environment.

        Parameters
        ----------
        action_space : Optional[Space]
            The action space for the ego agent. If not given a Box space is
            assumed for acceleration and steering actions.

        observation_space : Optional[Space]
            The observation space for the ego agent. If not given a Box space is
            assumed for a rasterized map.

        reward_range : Tuple[float, float]
            Optional reward range parameter for gym.

        terminal_conditions : Optional[List[Union[str, Callable[[State], bool]]]]
            Terminal condtiions for the scenario gym. If not given then max_length,
            ego_collision and ego_off_road are used.

        timestep : float
            Timestep for the scenario_gym.

        create_agent : Optional[Callable[[Scenario, Entity], Optional[Agent]]]
            Create agent function for the gym. Should return an agent for the
            ego entity. If not given then `ScenarioGym.create_agent` will be used.

        select_scenario : Optional[Callable[[], Union[Scenario, str]]]
            Function that selects the scenario to be run each time `reset()` is
            called. Takes just self as argument and should return either the xosc
            filepath or the scenario object. If not given then
            `ScenarioGym.select_scenario` will be used.

        """
        if terminal_conditions is None:
            terminal_conditions = ["max_length", "ego_collision", "ego_off_road"]
        super().__init__(
            terminal_conditions=terminal_conditions,
            timestep=timestep,
            **kwargs,
        )
        if action_space is None:
            action_space = Box(
                low=np.array([-5.0, -0.9]).astype(np.float32),
                high=np.array([5.0, 0.9]).astype(np.float32),
                shape=(2,),
            )
        if observation_space is None:
            observation_space = Box(
                low=np.float32(0.0),
                high=np.float32(1.0),
                shape=(2, 128, 128),
            )
        self.action_space = action_space
        self.observation_space = observation_space
        self.reward_range = reward_range
        if create_agent is not None:
            self.create_agent = create_agent
        if select_scenario is not None:
            self.select_scenario = MethodType(select_scenario, self)

    def on_reset(self) -> None:
        """Run just before the reset is executed."""
        pass

    def after_reset(self) -> None:
        """Run just after the reset is executed."""
        pass

    def reset(
        self,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[Dict] = None,
    ) -> Union[Observation, Tuple[Observation, Dict]]:
        """
        Reset the environment.

        Resets the state and computes the observation for the ego agent.

        Possible options:

        scenario : Union[Scenario, str]
            A chosen scenario object or filepath to the xosc to be used.

        """
        self.on_reset()

        if self._new_reset:
            super().reset(seed=seed)
        else:
            super().seed(seed)
        if (options is not None) and ("scenario" in options):
            s = options["scenario"]
        else:
            s = self.select_scenario()
        if s is not None:
            if isinstance(s, Scenario):
                self.set_scenario(s)
            else:
                self.load_scenario(s)
        elif self.state.scenario is None:
            raise ValueError("No scenario has been set.")
        else:
            self.reset_scenario()

        self.state.next_t = self.state.t + self.timestep
        ego_obs = self.ego_agent.sensor.step(self.state)

        self.after_reset()
        return (ego_obs, {}) if return_info else ego_obs

    def step(self, action: Action) -> Tuple[Observation, float, bool, Dict]:
        """
        Run one timestep of the environment.

        The action for the ego is processed and `step` is called on
        all other agents/entities. Then the state is processed and the
        reward is  for the

        Returns
        -------
        next_state : Observation
            Observation of the next state for the ego.

        reward : float
            The reward returned from the next state.

        done : bool
            Whether the next state is terminal.

        info : Dict
            Additional info.

        """
        if self.state.is_done:
            raise ValueError("Step called when state is terminal.")

        new_poses = {}
        for agent in self.state.agents.values():
            if agent is self.ego_agent:
                agent.last_action = action
                new_poses[agent.entity] = agent.controller.step(self.state, action)
            else:
                new_poses[agent.entity] = agent.step(self.state)
        new_poses.update(self.state.non_agents.step(self.state))

        # update the poses and current time
        self.state.step(new_poses)

        # get reward of next state
        reward = self.ego_agent.reward(self.state)

        # rendering and metrics
        if self.viewer is not None:
            self.state.last_keystroke = self.render()
        for m in self.metrics:
            m.step(self.state)

        # process ego part of next state
        self.state.next_t = self.state.t + self.timestep
        ego_obs = self.ego_agent.sensor.step(self.state)

        if self.state.is_done:
            for agent in self.state.agents.values():
                agent.finish(self.state)

        return ego_obs, reward, self.state.is_done, {}

    def rollout(self, *args, **kwargs):
        """Raise an error if rollout is called with this env."""
        raise NotImplementedError("Rollout is not supported for this environment.")

    def render(
        self,
        mode: None = None,
        video_path: Optional[str] = None,
    ) -> Optional[int]:
        """Render the environment."""
        return super().render(video_path=video_path)

    def load_scenario(
        self, *args, create_agent: Optional[Callable] = None, **kwargs
    ) -> None:
        """
        Load a scenario from an OpenScenario file.

        Sets the default argument of `create_agent` to `self.create_agent`.
        """
        if create_agent is None:
            create_agent = self.create_agent
        super().load_scenario(*args, create_agent=create_agent, **kwargs)

    def set_scenario(
        self, *args, create_agent: Optional[Callable] = None, **kwargs
    ) -> None:
        """
        Set the scenario explicitly.

        Sets the default argument of `create_agent` to `self.create_agent`.
        """
        if create_agent is None:
            create_agent = self.create_agent
        super().set_scenario(*args, create_agent=create_agent, **kwargs)

    def select_scenario(self) -> Optional[Union[str, Scenario]]:
        """Update the scenario when reset is called."""
        return None

    def create_agents(
        self,
        create_agent: Callable[[Scenario, Entity], Optional[Agent]] = _create_agent,
    ) -> None:
        """Check there is an ego agent."""
        super().create_agents(create_agent=create_agent)
        try:
            self.ego_agent = self.state.agents[self.state.scenario.ego]
        except KeyError as e:
            raise KeyError("No agent for ego.") from e

    @staticmethod
    def create_agent(scenario: Scenario, entity: Entity) -> Optional[Agent]:
        """Create the agents for the scenario."""
        if entity.ref == "ego":
            return RLAgent(
                entity,
                VehicleController(entity, max_steer=0.9, max_accel=5.0),
                MapOnlySensor(
                    entity, channels_first=True, height=30, width=30, n=128
                ),
            )


class MapOnlySensor(RasterizedMapSensor):
    """Sensor returning only the rasterized map."""

    def _step(self, state: State) -> np.ndarray:
        """Get the map from the base sensor's observation."""
        return super()._step(state).map


class RLAgent(Agent):
    """Example agent recieving negative rewards for collisions going off road."""

    def reward(self, state: State) -> Optional[float]:
        """Return the reward for the agent from the current state."""
        if state.is_done:
            if TERMINAL_CONDITIONS["ego_off_road"](state):
                return -1.0
            elif TERMINAL_CONDITIONS["ego_collision"](state):
                return -1.0
        return 0.01



from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from shapely.geometry import LinearRing, LineString, Polygon

from scenario_gym.entity import Entity, Pedestrian, Vehicle
from scenario_gym.state import State

from .base import Viewer
from .utils import to_ego_frame, vec2pix

# Blue, Green, Red color tuple
BGRTuple = Tuple[int, int, int]
# Blue, Green, Red, alpha color tuple
BGRATuple = Tuple[int, int, int, float]


@dataclass
class Color:
    """Store BGR color for OpenCV viewing with optional alpha channel."""

    B: int
    G: int
    R: int
    alpha: float = 1.0  # alpha \in [0.0, 1.0]

    @classmethod
    def create_from_tuple(cls, color_tuple: Union[BGRTuple, BGRATuple]):
        """
        Create an instance of the Color class from a BGR or BGRalpha tuple.

        Parameters
        ----------
        color_tuple : Union[BGRTuple, BGRATuple]
            Tuple of B, G, R values (between 0 and 255 inclusive) and an optional
            alpha (between 0.0 and 1.0 inclusive).

        Returns
        -------
        Color
            Instance of Color class corresponding to the color tuple passed in.

        Raises
        ------
        ValueError
            If the color tuple is not a valid color.

        """
        if not all(
            (
                isinstance(color_tuple, tuple),
                (len(color_tuple) == 3 or len(color_tuple) == 4),
                all(isinstance(x, int) for x in color_tuple[:3]),
                0 <= min(color_tuple[:3]),
                255 >= max(color_tuple[:3]),
            )
        ):
            raise ValueError(
                f"{color_tuple} is not a valid color. Must be a tuple of 3 ints "
                + "(and optionally 1 float for the alpha channel)."
            )
        if len(color_tuple) == 3:
            color = cls(*color_tuple)
        else:
            alpha = color_tuple[3]
            if not isinstance(alpha, float):
                raise ValueError(
                    f"Provided alpha value {alpha} is not valid; "
                    + "must be a float."
                )
            if alpha < 0.0 or alpha > 1.0:
                raise ValueError(
                    f"Provided alpha value {alpha} is not valid; "
                    + "must be between 0.0 and 1.0."
                )
            color = cls(*color_tuple[:3], alpha=alpha)

        return color

    @property
    def bgr(self) -> BGRTuple:
        """Return the BGR tuple corresponding to this color."""
        return (self.B, self.G, self.R)


class OpenCVViewer(Viewer):
    """
    General display class to visualise the Scenario Gym.

    Rendering behaviour can be customised by changing the layers that are rendered.
    The available layers and default colors are given by
    `OpenCVViewer._renderable_layers`.
    """

    _renderable_layers: Dict[str, Optional[Color]] = {
        "driveable_surface": Color(128, 128, 128),
        "driveable_surface_boundary": Color(250, 250, 250),
        "walkable_surface": Color(210, 210, 210),
        "buildings": Color(128, 128, 128),
        "roads": Color(128, 128, 128),
        "intersections": Color(128, 128, 128),
        "lanes": Color(128, 128, 128),
        "road_centers": Color(180, 180, 180),
        "lane_centers": Color(255, 128, 128),
        "text": Color(250, 250, 250),
    }

    def __init__(
        self,
        magnification: int = 10,
        fps: float = 30,
        headless_rendering: bool = True,
        render_entity: str = "ego",
        render_layers: Optional[List[str]] = None,
        codec: str = "avc1",
        line_thickness: int = 1,
        width: int = 100,
        height: int = 100,
        window_name: str = "frame",
        entity_color_dict: Optional[
            Dict[Entity, Union[Color, BGRTuple, BGRATuple]]
        ] = None,
        **colors: Union[Color, BGRTuple, BGRATuple],
    ):
        """
        Init the viewer.

        Parameters
        ----------
        magnification : int
            The number of pixels per meter.

        fps : float
            The frames per second.

        headless_rendering : bool
            Whether to render to an mp4 file or a popup window.

        output_path : Optional[str]
            The filepath if using headless rendering.

        render_entity : str
            The reference of the entity to use for centering the camera.

        render_layers : List[str]
            The layers to render. A list of all possible layers is given in
            `Viewer._renderable_layers`.

        codec : str
            Codec to encode videos e.g. mp4v or avc1.

        line_thickness : int
            Thickness of lines in pixels.

        width : int
            Width of the camera window in meters.

        height : int
            Height of the camera window in meters.

        window_name : str
            Name for the rendering window in non-headless mode.

        entity_color_dict : Optional[
            Dict[Entity, Union[Color, BGRTuple, BGRATuple]]
        ]
            Dictionary linking entities to colors in which they should be rendered.
            By default None, meaning all entities will use their default colors.

        colors : Union[Color, BGRTuple, BGRATuple]
            Color changes as keyword arguments.

        """
        super().__init__()
        self.headless_rendering = headless_rendering
        self.codec = codec
        self.w = width
        self.h = height
        self.mag = magnification
        self.fps = fps
        self.entity_ref = render_entity
        self.window_name = window_name
        self.line_thickness = line_thickness
        self.preset_entity_color_dict = (
            {} if entity_color_dict is None else entity_color_dict
        )
        self.centring_position = np.array([self.w / 2, self.h / 2])
        self.origin = np.array([0.0, 0.0])
        self._coords_cache = {}

        if render_layers is None:
            render_layers = [
                "driveable_surface_boundary",
                "road_centers",
                "lane_centers",
                "buildings",
            ]

        for layer in render_layers:
            if layer != "entity" and layer != "text":
                try:
                    getattr(self, f"render_{layer}")
                except AttributeError as e:
                    raise NotImplementedError(
                        f"Rendering method `render_{layer}` is not implemented."
                    ) from e
        self.render_layers = render_layers

        for k, v in {
            "background": (0, 0, 0),
            "entity_front": (250, 250, 250),
            **self._renderable_layers,
            **colors,
        }.items():
            if isinstance(v, tuple):
                v = Color.create_from_tuple(v)
            setattr(self, f"{k}_color", v)

        self.base_frame = (
            np.ones(
                [int(self.mag * self.w), int(self.mag * self.h), 3],
                dtype=np.uint8,
            )
            * np.array(self.background_color.bgr, dtype=np.uint8)[None, None, :]
        )
        self._frame = self.reset_frame()
        self._state = self._entity_colour_dict = None
        self.video_writer = None

    def reset_frame(self) -> np.ndarray:
        """Reset the frame to a black image."""
        return self.base_frame.copy()

    def render(self, state: State) -> Optional[int]:
        """
        Display the state of the gym at a given time.

        This will visualise a top-down view of the gym centrered around the
        ego agent.
        """
        self._state = state
        key = None
        self.draw_frame(state, e_ref=self.entity_ref)

        if self.headless_rendering:
            self.video_writer.write(self._frame)
        else:
            cv2.imshow(self.window_name, self._frame)
            wait_time = 0 if self.fps == 0 else 1000 // self.fps
            key = cv2.waitKey(wait_time)

        # reset the frame to delete current step
        self._frame = self.reset_frame()
        return key

    def reset(self, output_path: Optional[str] = None) -> None:
        """Reset at the start of a scenario rollout."""
        super().reset(output_path)
        self._frame = self.reset_frame()
        self._state = self._entity_colour_dict = None
        self.video_writer = None
        self._coords_cache = {}
        if self.headless_rendering:
            fourcc = cv2.VideoWriter_fourcc(*self.codec)
            self.video_writer = cv2.VideoWriter(
                output_path,
                fourcc,
                self.fps,
                (self.w * self.mag, self.h * self.mag),
            )

    def close(self) -> None:
        """Clean up any open files for the rendering."""
        cv2.destroyAllWindows()
        if self.video_writer:
            self.video_writer.release()

    def draw_frame(self, state: State, e_ref: Optional[str] = "ego") -> None:
        """Render the given state around a given entity (by reference)."""
        ego_pose = self.get_center_pose(state, e_ref)

        for layer in self.render_layers:
            if layer != "entity" and layer != "text":
                getattr(self, f"render_{layer}")(state, ego_pose)

        for entity, pose in state.poses.items():
            entity_idx = state.scenario.entities.index(entity)
            self.draw_entity(state, entity_idx, entity, pose, ego_pose)

        if "text" in self.render_layers:
            self.render_text(state)

    def get_center_pose(self, state: State, e_ref: str) -> np.ndarray:
        """Get the pose for the center of the frame."""
        ego_pose = None
        entity = state.scenario.entity_by_name(e_ref)
        if entity is None:
            entity = state.scenario.ego
        if entity in state.poses:
            ego_pose = state.poses[entity]
        else:
            ego_pose = state.recorded_poses(entity)
            ego_pose = (
                entity.trajectory[0][1:] if len(ego_pose) == 0 else ego_pose[0, 1:]
            )
        return ego_pose

    @property
    def entity_colors(self) -> Dict[int, Color]:
        """Store the entity color dict."""
        if self._entity_colour_dict is None:
            if self._state is None:
                raise ValueError(
                    "Cannot query the entity colors if state is not set."
                )
            self._entity_colour_dict = {
                i: self.get_entity_color(i, e)
                for i, e in enumerate(self._state.scenario.entities)
            }

        return self._entity_colour_dict

    def get_entity_color(self, entity_idx: int, entity: Entity) -> Color:
        """Get the color to draw the given entity."""
        if entity in self.preset_entity_color_dict.keys():
            c = self.preset_entity_color_dict[entity]
            if isinstance(c, tuple):
                c = Color.create_from_tuple(c)
        elif entity_idx == 0:
            c = Color(0, 0, 128)
        elif isinstance(entity, Vehicle):
            c = Color(128, 0, 0)
        elif isinstance(entity, Pedestrian):
            c = Color(0, 128, 0)
        else:
            c = Color(221, 160, 221)
        return c

    def draw_entity(
        self,
        state: State,
        entity_idx: int,
        entity: Entity,
        pose: np.ndarray,
        ego_pose: np.ndarray,
    ) -> None:
        """Draw the given entity onto the frame."""
        # get the bounding box in the global frame
        bbox_coords = entity.get_bounding_box_points(pose)

        # convert to the egos frame (rotated 90)
        bbox_in_ego = to_ego_frame(bbox_coords, ego_pose)

        # render front of the entity
        l = np.linalg.norm(bbox_in_ego[0] - bbox_in_ego[1])
        w = 0.25 / l
        front_bbox = np.array(
            [
                # 0.5 * (bbox_in_ego[2] + bbox_in_ego[1]),
                bbox_in_ego[2],
                bbox_in_ego[1],
                bbox_in_ego[1] + w * (bbox_in_ego[0] - bbox_in_ego[1]),
                bbox_in_ego[2] + w * (bbox_in_ego[3] - bbox_in_ego[2]),
            ]
        )

        # add to frame
        c = self.entity_colors[entity_idx]
        # Extra check that c is an instance of Color, for backwards compatibility
        # with subclasses of OpenCVViewer.
        if isinstance(c, tuple):
            c = Color.create_from_tuple(c)

        xy = vec2pix(bbox_in_ego, mag=self.mag, h=self.h, w=self.w)
        xy_front = vec2pix(front_bbox, mag=self.mag, h=self.h, w=self.w)

        entity_poly_frame = self._frame.copy()
        cv2.fillPoly(entity_poly_frame, [xy], c.bgr)
        cv2.fillPoly(entity_poly_frame, [xy_front], self.entity_front_color.bgr)

        self._frame = cv2.addWeighted(
            self._frame, (1.0 - c.alpha), entity_poly_frame, c.alpha, 0
        )

    def render_text(self, state: State) -> None:
        """Add text to the frame."""
        v = np.linalg.norm(state.velocities[state.scenario.ego][:3])
        cv2.putText(
            self._frame,
            "Ego speed: {:.2f}".format(v),
            (10, int(0.9 * self.mag * self.h)),  # bottom left corner of text
            cv2.FONT_HERSHEY_SIMPLEX,  # font
            1,  # font scale
            self.text_color.bgr,  # font color
            1,  # thickness
            2,  # line type
        )

    def draw_geom(
        self,
        geom: Union[Polygon, LineString, LinearRing],
        ego_pose: np.ndarray,
        c: Color,
        use_cache: bool = False,
    ) -> None:
        """Render a polygon or linestring to the frame."""
        if not isinstance(geom, (Polygon, LineString, LinearRing)):
            raise TypeError(f"{type(geom)} not supported.")

        xy = to_ego_frame(self.get_coords(geom, use_cache=use_cache), ego_pose)
        xy = vec2pix(xy, mag=self.mag, h=self.h, w=self.w)
        if isinstance(geom, LineString):
            cv2.polylines(self._frame, [xy], False, c.bgr, self.line_thickness)
        else:
            cv2.fillPoly(self._frame, [xy], c.bgr)
            for interior in geom.interiors:
                # cannot cache as the id is different each time
                xy = to_ego_frame(np.array(interior.xy).T, ego_pose)
                xy = vec2pix(xy, mag=self.mag, h=self.h, w=self.w)
                cv2.fillPoly(self._frame, [xy], self.background_color.bgr)

    def get_coords(
        self,
        geom: Union[Polygon, LineString, LinearRing],
        use_cache: bool = True,
    ) -> np.ndarray:
        """Get the coordinates for the given geometry."""
        if geom in self._coords_cache:
            return self._coords_cache[geom]
        coords = np.array(
            geom.exterior.xy if isinstance(geom, Polygon) else geom.xy
        ).T
        if use_cache:
            self._coords_cache[geom] = coords
        return coords

    def render_driveable_surface(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the driveable surface."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for geom in road_network.driveable_surface.geoms:
            self.draw_geom(
                geom, ego_pose, self.driveable_surface_color, use_cache=True
            )

    def render_driveable_surface_boundary(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the boundary of the driveable surface."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for geom in road_network.driveable_surface.geoms:
            self.draw_geom(
                geom.exterior,
                ego_pose,
                self.driveable_surface_boundary_color,
                use_cache=True,
            )
            for interior in geom.interiors:
                self.draw_geom(
                    interior,
                    ego_pose,
                    self.driveable_surface_boundary_color,
                    use_cache=True,
                )

    def render_walkable_surface(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the walkable surface."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for g in road_network.pavements + road_network.crossings:
            self.draw_geom(
                g.boundary, ego_pose, self.walkable_surface_color, use_cache=True
            )

    def render_buildings(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render any buildings."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for b in road_network.buildings:
            self.draw_geom(
                b.boundary, ego_pose, self.buildings_color, use_cache=True
            )

    def render_roads(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the road boundary polygons."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for r in road_network.roads:
            self.draw_geom(r.boundary, ego_pose, self.roads_color, use_cache=True)

    def render_road_centers(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the road centers."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for r in road_network.roads:
            self.draw_geom(
                r.center, ego_pose, self.road_centers_color, use_cache=True
            )

    def render_lanes(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the lane boudnary polygons."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for l in road_network.lanes:
            self.draw_geom(l.boundary, ego_pose, self.lanes_color, use_cache=True)

    def render_lane_centers(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the lane centers."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for l in road_network.lanes:
            self.draw_geom(
                l.center, ego_pose, self.lane_centers_color, use_cache=True
            )

    def render_intersections(
        self,
        state: State,
        ego_pose: np.ndarray,
    ) -> None:
        """Render the road boundary polygons."""
        road_network = state.scenario.road_network
        if road_network is None:
            return
        for i in road_network.intersections:
            self.draw_geom(
                i.boundary, ego_pose, self.intersections_color, use_cache=True
            )



import os
import pickle
from copy import deepcopy
from tempfile import TemporaryDirectory, TemporaryFile

import numpy as np
import pytest as pt

from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario import Scenario
from scenario_gym.scenario.actions import UpdateStateVariableAction
from scenario_gym.xosc_interface import import_scenario


@pt.fixture
def example_scenario(all_scenarios):
    """Get a scenario to test."""
    s = import_scenario(all_scenarios["3e39a079-5653-440c-bcbe-24dc9f6bf0e6"])
    action = UpdateStateVariableAction(2.0, "TestAction", "ego", {"var": 1.0})
    s.add_action(action, inplace=True)
    return s


@pt.fixture
def example_entity(example_scenario):
    """Get an entity to use as an example."""
    return deepcopy(example_scenario.entities[0])


def test_ego(example_scenario):
    """Test getting the ego entity."""
    assert example_scenario.ego is not None, "Ego entity should not be None."
    ego = example_scenario.entity_by_name("ego")
    first = example_scenario.entities[0]
    assert (
        ego == first == example_scenario.ego
    ), "Here the ego entity should be first and named 'ego'."

    other = [e.copy() for e in example_scenario.entities]
    other[0].ref = "not_ego"
    s = Scenario(other)
    assert s.ego == s.entities[0], "Ego should still be the first entity."

    other[1].ref = "ego"
    s2 = Scenario(other)
    assert s2.ego == s.entities[1], "Ego should now be the second entity."


def test_length(example_scenario):
    """Test the length of the scenario is computed correctly."""
    l = max((e.trajectory.max_t for e in example_scenario.entities))
    assert example_scenario.length == l, "Incorrect length."


def test_deepcopy_scenario(example_scenario):
    """Test deepcopying a scenario."""
    s_new = deepcopy(example_scenario)
    assert len(s_new.entities) == len(example_scenario.entities)
    assert id(s_new.road_network) != id(
        example_scenario.road_network
    ), "Road networks should be different."


def test_copy_scenarios(example_entity):
    """Test copying scenario."""
    s = Scenario(
        [example_entity],
        road_network=RoadNetwork(),
        properties={"x": 1, "y": 2},
    )
    s_new = s.copy()
    assert id(s_new.road_network) == id(
        s.road_network
    ), "Road networks should be the same."
    assert s_new.properties == s.properties, "Properties should be the same."


def test_add_entity(example_scenario, example_entity):
    """Test adding an entity to a scenario."""
    s = example_scenario.copy()
    new_entity = example_entity.copy()
    new_entity.ref = "new_entity"
    s.add_entity(new_entity, inplace=True)
    assert len(s.entities) == len(example_scenario.entities) + 1
    assert new_entity in s.entities, "Entity should be in the scenario."
    assert s.entity_by_name("new_entity") == new_entity

    s2 = example_scenario.add_entity(new_entity, inplace=False)
    assert len(s2.entities) == len(example_scenario.entities) + 1
    assert new_entity in s2.entities, "Entity should be in the scenario."
    assert s2.entity_by_name("new_entity") == new_entity
    assert new_entity not in example_scenario.entities


def test_remove_entity(example_scenario):
    """Test removing an entity from a scenario."""
    s = example_scenario.copy()
    e = s.entities[-1]
    s.remove_entity(e, inplace=True)
    assert len(s.entities) == len(example_scenario.entities) - 1
    assert e not in s.entities, "Entity should not be in the scenario."

    s2 = example_scenario.copy()
    e = s2.entities[-1]
    s3 = s2.remove_entity(e, inplace=False)
    assert len(s3.entities) == len(s2.entities) - 1
    assert e in s2.entities, "Entity should be in the scenario."
    assert e not in s3.entities, "Entity should not be in the new scenario."


def test_describe(example_scenario):
    """Test the `describe` method."""
    example_scenario.describe()


def test_plot(example_scenario):
    """Test the `plot` method."""
    example_scenario.plot(show=False)


def test_translate(example_scenario):
    """Test the translating a scenario."""
    shift = np.arange(7) + 1
    new_scenario = example_scenario.translate(shift)
    ts = np.linspace(0.0, example_scenario.length, 10)
    for e, e_new in zip(example_scenario.entities, new_scenario.entities):
        ps = e.trajectory.position_at_t(ts)
        ps_new = e_new.trajectory.position_at_t(ts + shift[0])
        deltas = ps_new - ps
        assert np.allclose(deltas, shift[None, 1:])

    a_new = new_scenario.actions[0]
    a_old = example_scenario.actions[0]
    assert a_new.t == a_old.t + shift[0]


def test_to_dict(example_scenario, all_scenarios):
    """Test writing and reading the scenario from a dictionary."""
    base_dir = os.path.join(
        os.path.dirname(all_scenarios[example_scenario.name]),
        "../Road_Networks",
    )
    data = example_scenario.to_dict(road_network_path=base_dir)
    s2 = Scenario.from_dict(data)
    assert (
        example_scenario.road_network.name == s2.road_network.name
    ), "Road networks should have same name."
    assert all(
        e.ref == e2.ref for e, e2 in zip(example_scenario.entities, s2.entities)
    ), "Entities should be in the same order."
    ego, ego2 = example_scenario.entities[0], s2.entities[0]
    assert (
        ego.trajectory.data == ego2.trajectory.data
    ).all(), "Ego trajectories should be the same."
    assert ego.bounding_box == ego2.bounding_box


def test_jsonable(example_scenario):
    """Test reading and writing scenarios from json."""
    with TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "example.json")
        example_scenario.to_json(path, road_network_path=None)
        s2 = Scenario.from_json(path)

    assert (
        example_scenario.road_network.name == s2.road_network.name
    ), "Road networks should have same name."
    assert all(
        e.ref == e2.ref for e, e2 in zip(example_scenario.entities, s2.entities)
    ), "Entities should be in the same order."
    ego, ego2 = example_scenario.entities[0], s2.entities[0]
    assert (
        ego.trajectory.data == ego2.trajectory.data
    ).all(), "Ego trajectories should be the same."


@pt.fixture()
def scenario_with_none_values(example_scenario):
    """Create a scenario with None values in one of the catalog entries."""
    s = deepcopy(example_scenario)
    s.entities[0].catalog_entry.front_axle.max_steering = None
    s.entities[0].catalog_entry.reat_axle = None
    return s


def test_jsonable_with_none(scenario_with_none_values):
    """Test reading and writing scenarios from json."""
    with TemporaryDirectory() as tmpdir:
        path = os.path.join(tmpdir, "example.json")
        scenario_with_none_values.to_json(path, road_network_path=None)
        s2 = Scenario.from_json(path)
    assert (
        s2.entities[0].catalog_entry.front_axle.max_steering is None
    ), "The max steering should be None."
    assert (
        scenario_with_none_values.road_network.name == s2.road_network.name
    ), "Road networks should have same name."
    assert all(
        e.ref == e2.ref
        for e, e2 in zip(
            scenario_with_none_values.entities,
            s2.entities,
        )
    ), "Entities should be in the same order."
    ego, ego2 = scenario_with_none_values.entities[0], s2.entities[0]
    assert (
        ego.trajectory.data == ego2.trajectory.data
    ).all(), "Ego trajectories should be the same."


def test_pickle_scenario(example_scenario):
    """Test pickling a scenario."""
    with TemporaryFile() as f:
        pickle.dump(example_scenario, f)
        f.seek(0)
        new_scenario = pickle.load(f)
    assert all(
        e_new.ref == e.ref
        for e_new, e in zip(new_scenario.entities, example_scenario.entities)
    ), "Different entities"
    assert new_scenario.length == example_scenario.length, "Different length."
    assert (
        new_scenario.road_network.name == example_scenario.road_network.name
    ), "Different road networks."



import os
import warnings
from contextlib import suppress
from typing import Dict, List, Optional, Type

import numpy as np
from lxml import etree
from lxml.etree import Element

from scenario_gym.entity import Entity, Pedestrian, Vehicle
from scenario_gym.road_network import RoadNetwork
from scenario_gym.scenario import Scenario, ScenarioAction
from scenario_gym.scenario.actions import UserDefinedAction
from scenario_gym.trajectory import Trajectory
from scenario_gym.utils import load_properties_from_xml

from .catalogs import load_object, read_catalog


def import_scenario(
    osc_file: str,
    relabel: bool = True,
    entity_types: Optional[List[Type[Entity]]] = None,
) -> Scenario:
    """
    Import a scenario from an OpenScenario file.

    Parameters
    ----------
    osc_file : str
        The filepath to the OpenScenario file.

    relabel : bool
        Whether to relabel entities after loading.

    entity_types : Optional[List[Type[Entity]]]
        Additional entity types to use when loading the scenario. Can be used to
        allow custom entities to be loaded from OpenSCENARIO.

    """
    if not os.path.exists(osc_file):
        raise FileNotFoundError

    cwd = os.path.dirname(osc_file)
    et = etree.parse(osc_file)
    osc_root = et.getroot()
    entities = {}

    # Read catalogs:
    catalogs: Dict[str, Dict[str, Entity]] = {}
    for catalog_location in osc_root.iterfind("CatalogLocations/"):
        rel_catalog_path = catalog_location.find("Directory").attrib["path"]
        if not os.path.isabs(rel_catalog_path):
            catalog_path = os.path.join(cwd, rel_catalog_path)
        else:
            catalog_path = rel_catalog_path
        for catalog_file in os.listdir(catalog_path):
            if catalog_file.endswith(".xosc"):
                catalog, entries = read_catalog(
                    os.path.join(catalog_path, catalog_file),
                    entity_types=entity_types,
                )
                catalogs[catalog.name] = entries

    # Import road network:
    rn_path = None
    scene_graph_file = osc_root.find("RoadNetwork/SceneGraphFile")
    if scene_graph_file is not None:
        rn_path = scene_graph_file.attrib["filepath"]
    else:
        logic_file = osc_root.find("RoadNetwork/LogicFile")
        if logic_file is not None:
            rn_path = logic_file.attrib["filepath"]

    road_network = None
    if rn_path is not None:
        if not os.path.isabs(rn_path):
            filepath = os.path.join(cwd, rn_path)
        else:
            filepath = rn_path
        extension = os.path.splitext(filepath)[1]
        if extension == "":
            filepath = f"{filepath}.json"
        with suppress(FileNotFoundError):
            road_network = RoadNetwork.create_from_file(filepath)

    # add the entities to the scenario
    for scenario_object in osc_root.iterfind("Entities/ScenarioObject"):
        entity_ref = scenario_object.attrib["name"]
        cat_ref = scenario_object.find("CatalogReference")
        if cat_ref is None:
            ent = None
            for element in scenario_object.getchildren():
                ent = load_object(element)
            if ent is None:
                warnings.warn(
                    "Could not find a catalog reference or entry for entity "
                    f"{ent.tag}.Perhaps you need to add an entity type to "
                    "`entity_types`."
                )
            else:
                ent.ref = entity_ref
                entities[entity_ref] = ent
        else:
            catalog_name = cat_ref.attrib["catalogName"]
            entry_name = cat_ref.attrib["entryName"]
            try:
                entity = catalogs[catalog_name][entry_name].copy()
                entity.ref = entity_ref
                entities[entity_ref] = entity
            except KeyError as e:
                if catalog_name not in catalogs:
                    warnings.warn(f"Could not find catalog: {catalog_name}")
                elif entry_name not in catalogs[catalog_name]:
                    warnings.warn(
                        f"Could not find entry {entry_name} in catalog "
                        f"{catalog_name}."
                    )
                else:
                    raise e

    # Read init actions:
    for private in osc_root.iterfind("Storyboard/Init/Actions/Private"):
        entity_ref = private.attrib["entityRef"]
        for wp in private.iterfind(
            "PrivateAction/TeleportAction/Position/WorldPosition"
        ):
            tp = traj_point_from_time_and_position(0, wp)
            # Add a single-waypoint trajectory:
            if entity_ref in entities:
                entities[entity_ref].trajectory = Trajectory(np.stack([tp], axis=0))

    # Read maneuver actions:
    actions = []
    for man_group in osc_root.iterfind("Storyboard/Story/Act/ManeuverGroup"):
        entity_ref = man_group.find("Actors/EntityRef")
        assert (
            entity_ref is not None
        ), "Could not find entity reference in maneuver group."
        entity_ref = entity_ref.attrib["entityRef"]
        entity = entities.get(entity_ref)

        if entity is None:
            continue

        for event in man_group.findall("Maneuver/Event"):
            traj_action = event.find(
                "Action/PrivateAction/RoutingAction/FollowTrajectoryAction"
            )
            if traj_action is not None:
                trajectory = read_trajectory_event(
                    traj_action,
                    road_network=road_network,
                )
                if trajectory is not None:
                    entity.trajectory = trajectory
                    continue

            user_action = event.find("Action/UserDefinedAction")
            start_trigger = event.find("StartTrigger")
            if user_action is not None:
                actions.extend(
                    load_user_defined_action(
                        entity,
                        user_action,
                        start_trigger=start_trigger,
                    )
                )

    header = osc_root.find("FileHeader")
    if header is not None:
        properties, files = load_properties_from_xml(header)
        if files and "files" not in properties:
            properties["files"] = files
    else:
        properties = {}

    scenario = Scenario(
        list(entities.values()),
        name=os.path.splitext(os.path.basename(osc_file))[0],
        road_network=road_network,
        properties=properties,
        actions=actions,
    )

    if relabel:
        scenario = relabel_scenario(scenario)

    return scenario


def read_trajectory_event(
    trajectory_action: Element,
    road_network: Optional[RoadNetwork] = None,
) -> Optional[Trajectory]:
    """Read a trajectory event from a ManeuverGroup."""
    # trajectory points
    trajectory_points = []
    vertices = trajectory_action.findall(
        "TrajectoryRef/Trajectory/Shape/Polyline/Vertex"
    )
    vertices.extend(trajectory_action.findall("Trajectory/Shape/Polyline/Vertex"))
    if not vertices:
        return None

    for vertex in vertices:
        t = float(vertex.attrib["time"])
        wp = vertex.find("Position/WorldPosition")
        trajectory_points.append(traj_point_from_time_and_position(t, wp))

    traj_data = np.stack(trajectory_points, axis=0)
    if (np.isnan(traj_data[:, 3]).sum() > 0) and (road_network is not None):
        traj_data[:, 3] = road_network.elevation_at_point(
            traj_data[:, 1], traj_data[:, 2]
        )

    return Trajectory(traj_data)


def load_user_defined_action(
    entity: Entity,
    user_action: Element,
    start_trigger: Optional[Element] = None,
) -> List[ScenarioAction]:
    """Load a user-defined action from an OpenSCENARIO file."""
    cond = start_trigger.find(
        "ConditionGroup/Condition/ByValueCondition/SimulationTimeCondition"
    )
    t = float(cond.attrib.get("value"))

    acts = []
    for child in user_action.findall("CustomCommandAction"):
        acts.append(
            UserDefinedAction(
                t,
                child.tag,
                entity.ref,
                {k: v for k, v in child.attrib.items()},
            )
        )
    return acts


def relabel_scenario(scenario: Scenario) -> Scenario:
    """
    Relabel the entities of the scenario.

    Will be relabelled to ego, vehicle_1, vehicle_2,
    ..., pedestrian_1, ..., other_1, ...

    """
    vehicles, pedestrians, others = 0, 0, 0
    scenario.entities[0].ref = "ego"
    old_to_new = {}
    for e in scenario.entities[1:]:
        cur = e.ref
        with suppress(KeyError):
            scenario._ref_to_entity.pop(cur)
        if isinstance(e, Vehicle):
            e.ref = f"vehicle_{vehicles}"
            vehicles += 1
        elif isinstance(e, Pedestrian):
            e.ref = f"pedestrian_{pedestrians}"
            pedestrians += 1
        else:
            e.ref = f"other_{others}"
            others += 1
        scenario._ref_to_entity[e.ref] = e
        old_to_new[cur] = e.ref
    for action in scenario.actions:
        if action.entity_ref in old_to_new:
            action.entity_ref = old_to_new[action.entity_ref]
    return scenario


def traj_point_from_time_and_position(t, world_position) -> np.ndarray:
    """Return the trajectory point as an array [t, x, y, z, h, p, r]."""
    return np.array(
        [
            t,
            float(world_position.attrib["x"]),
            float(world_position.attrib["y"]),
            float(world_position.attrib.get("z", np.NaN)),
            float(world_position.attrib.get("h", np.NaN)),
            float(world_position.attrib.get("p", np.NaN)),
            float(world_position.attrib.get("r", np.NaN)),
        ],
    )



from scenario_gym.callback import StateCallback
from scenario_gym.metrics import (
    CollisionMetric,
    EgoAvgSpeed,
    EgoDistanceTravelled,
    EgoMaxSpeed,
)
from scenario_gym.metrics.base import Metric, cache_mean
from scenario_gym.scenario_gym import ScenarioGym
from scenario_gym.state import State


def test_add_metric(all_scenarios):
    """Test adding a metric to the scenario."""
    s = all_scenarios["3fee6507-fd24-432f-b781-ca5676c834ef"]
    gym = ScenarioGym(
        metrics=[
            EgoAvgSpeed(),
            EgoMaxSpeed(),
            EgoDistanceTravelled(name="test_metric"),
            CollisionMetric(),
        ]
    )
    gym.load_scenario(s)
    gym.rollout()
    mets = [m.get_state() for m in gym.metrics]
    assert 4.0 <= mets[0] <= 5.0, "Avg speed incorrect"
    assert 10.0 <= mets[1] <= 12.0, "Max speed incorrect"
    assert 90.0 <= mets[2] <= 110.0, "Distance incorrect"
    assert not mets[3], "Collisions incorrect"

    data = gym.get_metrics()
    assert "test_metric" in data, "Name did not get applied."
    assert len(data) == 4, "Incrorect number of metrics returned."


def test_cache_mean(all_scenarios):
    """Test the cache_mean decorator."""
    s_ids = [
        "3fee6507-fd24-432f-b781-ca5676c834ef",
        "41dac6fa-6f83-461e-a145-08692da5f3c7",
    ]

    # get mean
    gym = ScenarioGym(
        metrics=[
            EgoAvgSpeed(),
        ]
    )

    gym.load_scenario(all_scenarios[s_ids[0]])
    gym.rollout()
    m1 = gym.metrics[0].get_state()

    gym.load_scenario(all_scenarios[s_ids[1]])
    gym.rollout()
    m2 = gym.metrics[0].get_state()

    avg = 0.5 * (m1 + m2)

    gym = ScenarioGym(
        metrics=[
            cache_mean(EgoAvgSpeed)(),
        ]
    )

    assert gym.metrics[0].previous_value == 0.0
    assert gym.metrics[0]._prev_count == 0.0

    gym.load_scenario(all_scenarios[s_ids[0]])
    gym.rollout()

    gym.load_scenario(all_scenarios[s_ids[1]])
    gym.rollout()

    assert gym.metrics[0].previous_value == avg
    assert gym.metrics[0].previous_value == 0.0


class ExampleCallback(StateCallback):
    """Callback for testing."""

    def _reset(self, state) -> None:
        """Set the value to 0."""
        self.value = 0.0

    def __call__(self, state: State) -> None:
        """Get the current time * 10."""
        self.value = state.t * 10.0


class DepMetric(Metric):
    """Example metric depending on a callback."""

    required_callbacks = [ExampleCallback]

    def _reset(self, state):
        """Set the value to the value from the callback."""
        self.value = self.callbacks[0].value

    def _step(self, state):
        """Set the value to the value from the callback / current time."""
        self.value = self.callbacks[0].value / state.t

    def get_state(self):
        """Return the value."""
        return self.value


def test_dependent_callback(all_scenarios):
    """Test using a metric that requires a callback."""
    s = all_scenarios["3fee6507-fd24-432f-b781-ca5676c834ef"]
    cb = ExampleCallback()
    met = DepMetric()
    gym = ScenarioGym(state_callbacks=[cb], metrics=[met])
    gym.load_scenario(s)
    assert met.callbacks == [cb]
    gym.rollout()
    assert cb.value == gym.state.t * 10.0
    assert met.get_state() == 10.0



from scenario_gym.entity.base import Entity
from scenario_gym.entity.batch import BatchReplayEntity
from scenario_gym.entity.misc import MiscObject, MiscObjectCatalogEntry
from scenario_gym.entity.pedestrian import Pedestrian, PedestrianCatalogEntry
from scenario_gym.entity.vehicle import Vehicle, VehicleCatalogEntry

DEFAULT_ENTITY_TYPES = (Vehicle, Pedestrian, MiscObject)



from abc import ABC, abstractmethod
from typing import Optional

from scenario_gym.state import State


class Viewer(ABC):
    """
    Base class for the scenario_gym visualization module.

    Subclasses will render a scenario rollout as a video. Each time
    a scenario is run a new instance of the viewer will be created
    which will render the rollout and then be destroyed.
    """

    def __init__(self):
        self.output_path: Optional[str] = None

    @abstractmethod
    def reset(self, output_path: str) -> None:
        """
        Reset the viewer at the start of a new scenario rollout.

        Parameters
        ----------
        output_path : str
            Path for the output video file.

        """
        self.output_path = output_path

    @abstractmethod
    def render(self, state: State) -> Optional[int]:
        """
        Render the current state of the environment.

        Should optionally return the current key press.

        Parameters
        ----------
        state : State
            Current global state of the gym.

        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the viewer."""
        raise NotImplementedError



import scenario_gym


def test_scenarios(all_scenarios):
    """Test all available scenarios."""
    gym = scenario_gym.ScenarioGym()
    for s in all_scenarios.values():
        gym.load_scenario(s)
        gym.rollout()



import os
from typing import Any, Callable, Dict, List, Optional, Type, Union

from scenario_gym.agent import Agent, _create_agent
from scenario_gym.entity import Entity
from scenario_gym.metrics import Metric
from scenario_gym.scenario import Scenario
from scenario_gym.state import State
from scenario_gym.viewer import Viewer
from scenario_gym.xosc_interface import import_scenario


class ScenarioGym:
    """The main class that loads and runs scenarios."""

    INIT_PREV_T = -0.1

    @classmethod
    def run_scenarios(
        cls,
        paths: List[str],
        render: bool = False,
        **kwargs,
    ) -> None:
        """Rollout the scenarios in paths."""
        gym = cls(**kwargs)
        for path in paths:
            gym.load_scenario(path)
            gym.rollout(render=render)

    def __init__(
        self,
        timestep: float = 1.0 / 30.0,
        persist: bool = False,
        viewer_class: Optional[Type[Viewer]] = None,
        terminal_conditions: Optional[
            List[Union[str, Callable[[State], bool]]]
        ] = None,
        state_callbacks: Optional[List[Callable[[State], None]]] = None,
        metrics: Optional[List[Metric]] = None,
        **viewer_parameters,
    ):
        """
        Init the gym.

        All arguments for the constructor of the viewer class should be
        passed as keyword arguments which will be stored.

        Parameters
        ----------
        timestep: float
            Time between steps in the gym.

        persist: bool
            If True then entities will persist for the entire scenario.

        viewer_class: Type[Viewer]
            Class type of the viewer that will be inisitalised.

        terminal_conditions : Optional[List[Union[str, Callable[[State], bool]]]]
            Conditions that if any are met will end the scenario.

        state_callbacks: Optional[List[Callable[[State], None]]]
            Additional methods to be called on the state after every step.

        metrics: List[Metric]
            List of metrics to measure.

        viewer_parameters:
            Keyword arguments for viewer_class.

        """
        self.timestep = timestep
        self.persist = persist
        if viewer_class is None and "fps" not in viewer_parameters:
            viewer_parameters["fps"] = int(1.0 / self.timestep)
        self.viewer_parameters = viewer_parameters.copy()

        if terminal_conditions is None:
            terminal_conditions = ["max_length"]
        self.terminal_conditions = terminal_conditions

        if state_callbacks is None:
            state_callbacks = []
        self.state_callbacks = state_callbacks

        if viewer_class is None:
            self._get_viewer()
        else:
            self.viewer_class = viewer_class
            self._render_enabled = True
        self.state: Optional[State] = None
        self.viewer: Optional[Viewer] = None
        self.reset_gym()

        if metrics is not None:
            self.add_metrics(metrics)

    def _get_viewer(self) -> None:
        """Get the viewer if it is not provided."""
        try:
            from scenario_gym.viewer.opencv import OpenCVViewer

            self.viewer_class = OpenCVViewer
            self._render_enabled = True
        except ImportError:
            self._render_enabled = False
            self.viewer_class = None

    def reset_gym(self) -> None:
        """
        Reset the state of the gym.

        Closes the viewer, removes any metrics and unloads the scenario.
        """
        self.close()
        self.state = None
        self.metrics = []

    def add_metrics(self, metrics: List[Metric]) -> None:
        """Add metrics to the gym."""
        self.metrics.extend(metrics)

    def load_scenario(
        self,
        scenario_path: str,
        create_agent: Callable[[Scenario, Entity], Optional[Agent]] = _create_agent,
        relabel: bool = False,
        **kwargs,
    ) -> None:
        """
        Load a scenario from a file.

        Parameters
        ----------
        scenario_path : str
            The scenario file to be loaded. Can be OpenSCENARIO or JSON.

        create_agent : Callable[[str, Entity], Optional[Agent]]
            A function that returns an agent to control a given entity.

        relabel : bool
            If given, all entities will be relabeled to ego, vehicle_1,
            vehicle_2, ..., pedestrian_1, ..., entity_1, ...

        """
        if scenario_path.endswith(".json"):
            scenario = Scenario.from_json(scenario_path, **kwargs)
        else:
            scenario = import_scenario(
                scenario_path,
                relabel=relabel,
                **kwargs,
            )
        self.set_scenario(
            scenario, scenario_path=scenario_path, create_agent=create_agent
        )

    def set_scenario(
        self,
        scenario: Scenario,
        scenario_path: Optional[str] = None,
        create_agent: Callable[[Scenario, Entity], Optional[Agent]] = _create_agent,
    ) -> None:
        """
        Update the current scenario and create agents.

        Parameters
        ----------
        scenario : Scenario
            The scenario object.

        scenario_path : Optional[str]
            The path to the scenario file if it was loaded from one.

        create_agent : Callable[[str, Entity], Optional[Agent]]
            A function that returns an agent to control a given entity.

        """
        self.state = State(
            scenario,
            scenario_path=scenario_path,
            persist=self.persist,
            conditions=self.terminal_conditions,
            state_callbacks=self.state_callbacks,
        )
        self.create_agents(create_agent=create_agent)
        self.reset_scenario()

    def create_agents(
        self,
        create_agent: Callable[[Scenario, Entity], Optional[Agent]] = _create_agent,
    ) -> None:
        """
        Create the agents for the scenario.

        Parameters
        ----------
        create_agent : Callable[[str, Entity], Optional[Agent]]
            A function that will return an agent for the given entity or
            return None if that entity does not need an agent. All entities
            without agents will replay their trajectories.

        """
        non_agents, non_agent_trajs = [], []
        for entity in self.state.scenario.entities:
            agent = create_agent(self.state.scenario, entity)
            if agent is not None:
                self.state.agents[entity] = agent
            else:
                non_agents.append(entity)
                non_agent_trajs.append(entity.trajectory)
        self.state.non_agents.add_entities(non_agents, non_agent_trajs)

    def get_start_time(self, scenario: Scenario) -> float:
        """Get the start time of the scenario."""
        return max((0.0, scenario.ego.trajectory.min_t))

    def reset_scenario(self) -> None:
        """Reset the state to the beginning of the current scenario."""
        self.close()
        t0 = self.get_start_time(self.state.scenario)
        self.state.reset(t0 + self.INIT_PREV_T, t0)
        for m in self.metrics:
            m.reset(self.state)

    def step(self) -> None:
        """Process a single step in the environment."""
        self.state.next_t = self.state.t + self.timestep

        # get the new poses
        new_poses = {}
        for entity, agent in self.state.agents.items():
            if entity in self.state.poses:
                pose = agent.step(self.state)
                if pose is not None:
                    new_poses[entity] = pose
                elif self.persist:
                    new_poses[entity] = self.state.poses[entity]
            elif entity.trajectory.min_t >= self.state.t:
                # the agent is initialised at its start position
                new_poses[entity] = entity.trajectory.position_at_t(
                    self.state.next_t
                )
        new_poses.update(self.state.non_agents.step(self.state))

        # update the poses and current time
        self.state.step(new_poses)

        # metrics and rendering
        for m in self.metrics:
            m.step(self.state)
        if self.viewer is not None:
            self.state.last_keystroke = self.render()

    def rollout(
        self, render: bool = False, video_path: Optional[str] = None
    ) -> None:
        """Rollout the current scenario fully."""
        self.reset_scenario()
        if render:
            self.state.last_keystroke = self.render(video_path=video_path)
        while not self.state.is_done:
            self.step()
        for agent in self.state.agents.values():
            agent.finish(self.state)
        self.close()

    def render(self, video_path: Optional[str] = None) -> None:
        """Render the state of the gym."""
        if self.viewer is None:
            self.reset_viewer(video_path=video_path)
        return self.viewer.render(self.state)

    def reset_viewer(self, video_path: Optional[str] = None) -> None:
        """Reset the viewer at the start of a new rollout."""
        if self.viewer is None:
            if not self._render_enabled:
                raise ValueError(
                    "Rendering is disabled since no `viewer_class` was provided "
                    "and the default viewer could not be imported. Perhaps OpenCV "
                    "is not installed?"
                )
            self.viewer = self.viewer_class(**self.viewer_parameters)
        else:
            self.viewer.close()
        if video_path is None:
            path = self.state.scenario_path
            video_dir = os.path.join(os.path.dirname(path), "../Recordings")
            if os.path.exists(video_dir):
                video_path = os.path.join(
                    video_dir,
                    os.path.splitext(
                        os.path.basename(path),
                    )[0]
                    + ".mp4",
                )
            else:
                video_path = os.path.splitext(self.state.scenario_path)[0] + ".mp4"
        self.viewer.reset(video_path)

    def close(self) -> None:
        """Close the gym."""
        if self.viewer is not None:
            self.viewer.close()
            self.viewer = None

    def get_metrics(self) -> Dict[str, Any]:
        """Get the current metric states."""
        values = {}
        for metric in self.metrics:
            value = metric.get_state()
            if isinstance(value, dict):
                for k, v in value.items():
                    if isinstance(k, str):
                        values[f"{metric.name}_{k}"] = v
            elif value is not None:
                values[metric.name] = value
        return values



import numpy as np

from scenario_gym import ScenarioGym
from scenario_gym.agent import PIDAgent


def test_pid_controller(all_scenarios):
    """Test the PID controller."""
    scenario_path = all_scenarios["a98d5c7d-76aa-49bf-b88c-97db5d5c7433"]
    gym = ScenarioGym(timestep=0.1)

    def create_agent(s, e):
        if e.ref == "ego":
            return PIDAgent(
                e,
                accel_Kp=2.0,
                max_accel=5.0,
                max_steer=np.pi / 90,
            )

    gym.load_scenario(
        scenario_path,
        create_agent=create_agent,
    )
    gym.rollout(render=False)
